from __future__ import annotations

# standard libraries
import contextlib
import datetime
import gettext
import json
import logging
import os
import pathlib
import typing
import uuid

# local libraries
from nion.swift.model import Cache
from nion.swift.model import Changes
from nion.swift.model import DocumentModel
from nion.swift.model import FileStorageSystem
from nion.swift.model import Observer
from nion.swift.model import Persistence
from nion.swift.model import Project
from nion.utils import Converter
from nion.utils import Observable


_ = gettext.gettext


ProfileContext = typing.TypeVar("ProfileContext")


class ProjectReference(Observable.Observable, Persistence.PersistentObject):

    def __init__(self, type: str):
        super().__init__()
        self.define_type(type)
        self.define_property("project_uuid", converter=Converter.UuidToStringConverter())
        self.define_property("last_used", None, converter=Converter.DatetimeToStringConverter())
        self.__has_project_info_been_read = False
        self.__project_version = None
        self.__project_state = "invalid"
        self.__document_model: typing.Optional[DocumentModel.DocumentModel] = None
        self.__document_model_about_to_close_listener = None
        self.storage_cache = None

    def close(self) -> None:
        if self.__document_model_about_to_close_listener:
            self.__document_model_about_to_close_listener.close()
            self.__document_model_about_to_close_listener = None
        super().close()

    def about_to_be_removed(self, container):
        self.unload_project()
        super().about_to_be_removed(container)

    def close_relationships(self) -> None:
        self.unload_project()
        super().close_relationships()

    def insert_model_item(self, container, name, before_index, item):
        """Insert a model item. Let this item's container do it if possible; otherwise do it directly.

        Passing responsibility to this item's container allows the library to easily track dependencies.
        However, if this item isn't yet in the library hierarchy, then do the operation directly.
        """
        if self.container:
            self.container.insert_model_item(container, name, before_index, item)
        else:
            container.insert_item(name, before_index, item)

    def remove_model_item(self, container, name, item, *, safe: bool=False) -> Changes.UndeleteLog:
        """Remove a model item. Let this item's container do it if possible; otherwise do it directly.

        Passing responsibility to this item's container allows the library to easily track dependencies.
        However, if this item isn't yet in the library hierarchy, then do the operation directly.
        """
        if self.container:
            return self.container.remove_model_item(container, name, item, safe=safe)
        else:
            container.remove_item(name, item)
            return Changes.UndeleteLog()

    def read_from_dict(self, properties):
        super().read_from_dict(properties)
        if not self.last_used:
            self.last_used = self._get_last_used()

    def _get_last_used(self) -> datetime.datetime:
        # fallback when last_used is not initialized
        return self.modified

    def __property_changed(self, name, value):
        self.notify_property_changed(name)

    @property
    def project_version(self) -> typing.Optional[int]:
        return self.project.project_version if self.project else self.__project_version

    @property
    def project_state(self) -> str:
        return self.project.project_state if self.project else self.__project_state

    @property
    def project(self) -> typing.Optional[Project.Project]:
        return self.__document_model._project if self.__document_model else None

    @property
    def document_model(self) -> typing.Optional[DocumentModel.DocumentModel]:
        return self.__document_model

    @property
    def title(self) -> str:
        if self.project and self.project.title:
            return self.project.title
        project_reference_parts = self.project_reference_parts
        if project_reference_parts:
            return pathlib.Path(project_reference_parts[-1]).stem
        return _("Untited")

    @property
    def project_reference_parts(self) -> typing.Tuple[str]:
        raise NotImplementedError()

    def make_storage(self, profile_context: typing.Optional[ProfileContext]) -> typing.Optional[FileStorageSystem.ProjectStorageSystem]:
        raise NotImplementedError()

    def read_project_info(self, profile_context: typing.Optional[ProfileContext]) -> None:
        if not self.__has_project_info_been_read:
            try:
                project_storage_system = self.make_storage(profile_context)
            except Exception:
                project_storage_system = None
            if project_storage_system:
                project_storage_system.load_properties()
                with contextlib.closing(Project.Project(project_storage_system)) as project:
                    if self.project_uuid != project.project_uuid:
                        self.project_uuid = project.project_uuid
                    self.__project_version = project.project_version
                    self.__project_state = project.project_state
                    self.__has_project_info_been_read = True

    def load_project(self, profile_context: typing.Optional[ProfileContext]) -> None:
        """Read project.

        The profile context is used during testing.
        """
        if not self.project:  # the project from the document model
            project: typing.Optional[Project.Project] = None

            # create project if it doesn't exist
            project_storage_system = self.make_storage(profile_context)
            if project_storage_system:
                project_storage_system.load_properties()
                project = Project.Project(project_storage_system)

            if project:
                self.__document_model = DocumentModel.DocumentModel(project, storage_cache=self.storage_cache)

                # handle special case of document model closing during tests
                def document_window_close():
                    self.__document_model_about_to_close_listener.close()
                    self.__document_model_about_to_close_listener = None
                    self.__document_model = None

                self.project.prepare_read_project()  # sets up the uuid, used next.

                self.update_item_context(self.project)
                self.project.about_to_be_inserted(self)
                self.notify_property_changed("project")  # before reading, so document model has a chance to set up
                self.project.read_project()

                self.__document_model_about_to_close_listener = self.__document_model.about_to_close_event.listen(document_window_close)
            else:
                logging.getLogger("loader").warning(f"Project could not be loaded {self}.")

    def unload_project(self) -> None:
        """Unload project (high level, notify that project changed)."""
        if self.project:
            self.project.unmount()
            self.project.about_to_be_removed(self)
            self.project.persistent_object_context = None
            self.__document_model.close()
            self.__document_model = None
            self.notify_property_changed("project")

    def read_project_uuid(self, profile_context: typing.Optional[ProfileContext] = None) -> typing.Optional[uuid.UUID]:
        """Read the project UUID without loading entire project."""
        project_storage_system = self.make_storage(profile_context)
        if project_storage_system:
            project_storage_system.load_properties()
            with contextlib.closing(Project.Project(project_storage_system)) as project:
                project.prepare_read_project()  # sets up the uuid, used next.
                return project.uuid
        return None

    def upgrade(self, profile_context: typing.Optional[ProfileContext] = None) -> typing.Optional[ProjectReference]:
        if self.project_state == "needs_upgrade":
            project_storage_system = self.make_storage(profile_context)
            if project_storage_system:
                return self._upgrade_project_storage_system(project_storage_system)
        return None

    def _upgrade_project_storage_system(self, project_storage_system: FileStorageSystem.ProjectStorageSystem) -> ProjectReference:
        raise NotImplementedError()


class IndexProjectReference(ProjectReference):
    type = "project_index"

    def __init__(self):
        super().__init__(self.__class__.type)
        self.define_property("project_path", converter=Converter.PathToStringConverter())

    @property
    def project_reference_parts(self) -> typing.Tuple[str]:
        return self.project_path.parts if self.project_path else tuple()

    def make_storage(self, profile_context: typing.Optional[ProfileContext]) -> typing.Optional[FileStorageSystem.ProjectStorageSystem]:
        return FileStorageSystem.make_index_project_storage_system(self.project_path)

    def _get_last_used(self) -> datetime.datetime:
        # fallback when last_used is not initialized
        try:
            return datetime.datetime.fromtimestamp(os.path.getmtime(self.project_path))
        except Exception:
            return super()._get_last_used()


class FolderProjectReference(ProjectReference):
    type = "project_folder"

    def __init__(self):
        super().__init__(self.__class__.type)
        self.define_property("project_folder_path", converter=Converter.PathToStringConverter())

    @property
    def project_reference_parts(self) -> typing.Tuple[str]:
        return self.project_folder_path.parts if self.project_folder_path else tuple()

    def make_storage(self, profile_context: typing.Optional[ProfileContext]) -> typing.Optional[FileStorageSystem.ProjectStorageSystem]:
        if self.project_folder_path:
            return FileStorageSystem.make_folder_project_storage_system(self.project_folder_path)
        return None

    def _get_last_used(self) -> datetime.datetime:
        # fallback when last_used is not initialized
        try:
            return datetime.datetime.fromtimestamp(os.path.getmtime(self.project_folder_path))
        except Exception:
            return super()._get_last_used()

    def _upgrade_project_storage_system(self, project_storage_system: FileStorageSystem.ProjectStorageSystem) -> ProjectReference:
        legacy_path = pathlib.Path(project_storage_system.get_identifier())
        target_project_path = legacy_path.parent.with_suffix(".nsproj")
        target_data_path = target_project_path.with_name(target_project_path.stem + " Data")
        if target_project_path.exists() or target_data_path.exists():
            raise FileExistsError()
        logging.getLogger("loader").info(f"Created new project {target_project_path} {target_data_path}")
        target_project_uuid = uuid.uuid4()
        target_project_data_json = json.dumps(
            {"version": FileStorageSystem.PROJECT_VERSION, "uuid": str(target_project_uuid),
             "project_data_folders": [str(target_data_path.stem)]})
        target_project_path.write_text(target_project_data_json, "utf-8")
        with contextlib.closing(FileStorageSystem.FileProjectStorageSystem(target_project_path)) as new_storage_system:
            new_storage_system.load_properties()
            FileStorageSystem.migrate_to_latest(project_storage_system, new_storage_system)
        new_project_reference = IndexProjectReference()
        new_project_reference.project_path = target_project_path
        new_project_reference.project_uuid = target_project_uuid
        return new_project_reference


project_reference_factory_hook = None

def project_reference_factory(lookup_id: typing.Callable[[str], str]) -> typing.Optional[ProjectReference]:
    type = lookup_id("type")
    if type == IndexProjectReference.type:
        return IndexProjectReference()
    if type == FolderProjectReference.type:
        return FolderProjectReference()
    if callable(project_reference_factory_hook):
        return project_reference_factory_hook(type)
    return None


class Profile(Observable.Observable, Persistence.PersistentObject):
    count = 0  # useful for detecting leaks in tests

    def __init__(self, storage_system=None, storage_cache=None, *, profile_context: typing.Optional[ProfileContext] = None):
        super().__init__()
        self.__class__.count += 1

        self.define_root_context()
        self.define_type("profile")
        self.define_relationship("project_references", project_reference_factory, insert=self.__insert_project_reference, remove=self.__remove_project_reference)
        self.define_property("last_project_reference", converter=Converter.UuidToStringConverter())
        self.define_property("work_project_reference_uuid", converter=Converter.UuidToStringConverter())
        self.define_property("closed_items", list())

        self.storage_system = storage_system or FileStorageSystem.MemoryPersistentStorageSystem()
        self.storage_system.load_properties()

        self.storage_cache = storage_cache or Cache.DictStorageCache()  # need to deallocate
        self.set_storage_system(self.storage_system)

        self.profile_context = None

        # helper object to produce the projects sequence
        oo = Observer.ObserverBuilder()
        oo.source(self).ordered_sequence_from_array("project_references").map(oo.x.prop("project")).filter(lambda x: x is not None).trampoline(self, "projects")
        self.__projects_observer = oo.make_observable()

        if profile_context:
            self.profile_context = profile_context
            self.add_project_memory()

    def close(self) -> None:
        self.storage_cache.close()
        self.storage_cache = None
        self.storage_system.close()
        self.storage_system = None
        self.__projects_observer.close()
        self.__projects_observer = None
        self.profile_context = None
        self.__class__.count -= 1
        super().close()

    def read_from_dict(self, properties):
        # cleanup from beta versions
        properties.pop("data_groups", None)
        properties.pop("data_item_references", None)
        properties.pop("data_item_variables", None)
        properties.pop("project_uuid", None)
        properties.pop("target_project_reference_uuid", None)
        properties.pop("workspace_uuid", None)
        properties.pop("workspaces", None)
        super().read_from_dict(properties)

    def __property_changed(self, name, value):
        self.notify_property_changed(name)

    def __insert_project_reference(self, name: str, before_index: int, project_reference: ProjectReference) -> None:
        project_reference.storage_cache = self.storage_cache
        project_reference.read_project_info(self.profile_context)
        self.notify_insert_item("project_references", project_reference, before_index)

    def __remove_project_reference(self, name: str, index: int, project_reference: ProjectReference) -> None:
        project_reference.storage_cache = None
        self.notify_remove_item("project_references", project_reference, index)

    @property
    def projects(self) -> typing.List[Project.Project]:
        return typing.cast(typing.List[Project.Project], self.__projects_observer.item)

    @property
    def _profile_storage_system(self) -> FileStorageSystem.PersistentStorageSystem:
        return self.storage_system

    def insert_model_item(self, container, name, before_index, item):
        """Insert a model item. Let this item's container do it if possible; otherwise do it directly.

        Passing responsibility to this item's container allows the library to easily track dependencies.
        However, if this item isn't yet in the library hierarchy, then do the operation directly.
        """
        if self.container:
            self.container.insert_model_item(container, name, before_index, item)
        else:
            container.insert_item(name, before_index, item)

    def remove_model_item(self, container, name, item, *, safe: bool=False) -> Changes.UndeleteLog:
        """Remove a model item. Let this item's container do it if possible; otherwise do it directly.

        Passing responsibility to this item's container allows the library to easily track dependencies.
        However, if this item isn't yet in the library hierarchy, then do the operation directly.
        """
        if self.container:
            return self.container.remove_model_item(container, name, item, safe=safe)
        else:
            container.remove_item(name, item)
            return Changes.UndeleteLog()

    def transaction_context(self):
        """Return a context object for a document-wide transaction."""

        class Transaction:
            def __init__(self, profile):
                self.__profile = profile

            def __enter__(self):
                self.__profile._profile_storage_system.enter_write_delay(self.__profile)
                for project in self.__profile.projects:
                    project.project_storage_system.enter_transaction()
                return self

            def __exit__(self, type, value, traceback):
                self.__profile._profile_storage_system.exit_write_delay(self.__profile)
                self.__profile._profile_storage_system.rewrite_item(self.__profile)
                for project in self.__profile.projects:
                    project.project_storage_system.exit_transaction()

        return Transaction(self)

    def read_profile(self) -> None:
        # read the properties from the storage system. called after open.
        properties = self.storage_system.read_properties()

        # if the properties match the current version, read the properties.
        if properties.get("version", 0) == FileStorageSystem.PROFILE_VERSION:
            self.begin_reading()
            try:
                if not self.project_references:  # hack for testing. tests will have already set up profile.
                    self.read_from_dict(properties)
            finally:
                self.finish_reading()
            self.storage_system.set_property(self, "uuid", str(self.uuid))
            self.storage_system.set_property(self, "version", FileStorageSystem.PROFILE_VERSION)

    def read_project(self, project_reference: ProjectReference) -> None:
        project_reference.load_project(self.profile_context)

    def create_project(self, project_dir: pathlib.Path, library_name: str) -> typing.Optional[ProjectReference]:
        project_name = pathlib.Path(library_name)
        project_data_path = pathlib.Path(library_name + " Data")
        project_path = project_dir / project_name.with_suffix(".nsproj")
        project_dir.mkdir(parents=True, exist_ok=True)
        project_uuid = uuid.uuid4()
        project_data_json = json.dumps({"version": FileStorageSystem.PROJECT_VERSION, "uuid": str(project_uuid), "project_data_folders": [str(project_data_path)]})
        project_path.write_text(project_data_json, "utf-8")
        project_reference = IndexProjectReference()
        project_reference.project_path = project_path
        project_reference.project_uuid = project_uuid
        self.append_project_reference(project_reference)
        return project_reference

    def open_project(self, path: pathlib.Path) -> typing.Optional[ProjectReference]:
        if path.suffix == ".nslib":
            return self.add_project_folder(pathlib.Path(path.parent), load=False)
        elif path.suffix == ".nsproj":
            return self.add_project_index(path, load=False)
        return None

    def get_project_reference(self, uuid_: uuid.UUID) -> typing.Optional[ProjectReference]:
        for project_reference in self.project_references:
            if project_reference.uuid == uuid_:
                return project_reference
        return None

    def append_project_reference(self, project_reference: ProjectReference) -> None:
        assert not self.get_item_by_uuid("project_references", project_reference.uuid)
        assert not project_reference.project_uuid in {project_reference.project_uuid for project_reference in self.project_references}
        self.append_item("project_references", project_reference)

    def remove_project_reference(self, project_reference: ProjectReference) -> None:
        assert project_reference.project_state != "loaded"
        project_reference.unload_project()
        self.remove_item("project_references", project_reference)

    def add_project_reference(self, project_reference: ProjectReference, load: bool = True) -> ProjectReference:
        # add the project reference if a project reference with the same project uuid
        # is not already present; otherwise activate the existing one.
        existing_project_reference = next(filter(lambda x: x.project_uuid == project_reference.project_uuid, self.project_references), None)
        if not existing_project_reference:
            self.append_project_reference(project_reference)
            if load:
                self.read_project(project_reference)
            return project_reference
        else:
            if load:
                existing_project_reference.load_project(self.profile_context)
            return existing_project_reference

    def add_project_index(self, project_path: pathlib.Path, load: bool = True) -> ProjectReference:
        project_reference = IndexProjectReference()
        project_reference.project_path = project_path
        project_reference.project_uuid = project_reference.read_project_uuid(self.profile_context)
        return self.add_project_reference(project_reference, load)

    def add_project_folder(self, project_folder_path: pathlib.Path, load: bool = True) -> ProjectReference:
        project_reference = FolderProjectReference()
        project_reference.project_folder_path = project_folder_path
        project_reference.project_uuid = project_reference.read_project_uuid(self.profile_context)
        return self.add_project_reference(project_reference, load)

    def add_project_memory(self, _uuid: uuid.UUID = None, load: bool = True) -> ProjectReference:
        assert callable(project_reference_factory_hook)
        project_reference = project_reference_factory_hook("project_memory")
        project_reference.project_uuid = _uuid or uuid.uuid4()
        return self.add_project_reference(project_reference, load)

    def upgrade(self, project_reference: ProjectReference) -> typing.Optional[ProjectReference]:
        new_project_reference = project_reference.upgrade(self.profile_context)
        if new_project_reference:
            self.remove_project_reference(project_reference)
            return self.add_project_reference(new_project_reference, load=False)
        return None



"""
    A module for handle .h5 files for Swift.
"""

import io
import json
import os
import pathlib
import threading
import typing

import h5py
import numpy

from nion.swift.model import StorageHandler
from nion.swift.model import Utility
from nion.utils import Geometry


def make_directory_if_needed(directory_path):
    """
        Make the directory path, if needed.
    """
    if os.path.exists(directory_path):
        if not os.path.isdir(directory_path):
            raise OSError("Path is not a directory:", directory_path)
    else:
        os.makedirs(directory_path)


def get_write_chunk_shape_for_data(data_shape, data_dtype):
    """
    Calculate an appropriate write chunk shape for a given data shape and dtype.

    The target chunk size is 580 kB which seems to be a sweet spot according to benchmarks.
    The algorithm assumes that the data is c-contiguous in memory.

    If the total number of chunks that the calculated chunk shape would lead to is less than 100 (i.e. the file will
    be less than 58 MB in size) or if the data shape is not suitable for chunking, return None.
    """
    data_dtype = numpy.dtype(data_dtype)

    target_chunk_size = 580*1024/data_dtype.itemsize
    chunk_size = 1
    counter = len(data_shape)
    chunk_shape = [1] * len(data_shape)
    while chunk_size < target_chunk_size and counter > 0:
        counter -= 1
        chunk_size *= data_shape[counter]
        chunk_shape[counter] = data_shape[counter]

    if chunk_size == 0: # This means one of the input dimensions was "0", so chunking cannot be used
        return None

    chunk_size /= data_shape[counter]
    remaining_elements = min(max(target_chunk_size // chunk_size, 1), data_shape[counter])
    chunk_shape[counter] = int(remaining_elements)

    n_chunks = 1
    for i in range(len(chunk_shape)):
        n_chunks *= data_shape[i] / chunk_shape[i]
    if n_chunks < 100:
        return None

    return tuple(chunk_shape)



class HDF5Handler(StorageHandler.StorageHandler):
    count = 0  # useful for detecting leaks in tests

    def __init__(self, file_path):
        self.__file_path = str(file_path)
        self.__lock = threading.RLock()
        self.__fp = None
        self.__dataset = None
        self._write_count = 0
        HDF5Handler.count += 1

    def close(self):
        HDF5Handler.count -= 1
        if self.__fp:
            self.__dataset = None
            self.__fp.close()
            self.__fp = None

    # called before the file is moved; close but don't count.
    def prepare_move(self) -> None:
        if self.__fp:
            self.__dataset = None
            self.__fp.close()
            self.__fp = None

    @property
    def reference(self):
        return self.__file_path

    @property
    def is_valid(self):
        return True

    @classmethod
    def is_matching(self, file_path):
        if file_path.endswith(".h5") and os.path.exists(file_path):
            return True
        return False

    @classmethod
    def make(cls, file_path: pathlib.Path):
        return cls(cls.make_path(file_path))

    @classmethod
    def make_path(cls, file_path: pathlib.Path) -> str:
        return str(file_path.with_suffix(cls.get_extension()))

    @classmethod
    def get_extension(self) -> str:
        return ".h5"

    def __ensure_open(self):
        if not self.__fp:
            make_directory_if_needed(os.path.dirname(self.__file_path))
            self.__fp = h5py.File(self.__file_path, "a")

    def __write_properties_to_dataset(self, properties):
        with self.__lock:
            assert self.__dataset is not None

            class JSONEncoder(json.JSONEncoder):
                def default(self, obj):
                    if isinstance(obj, Geometry.IntPoint) or isinstance(obj, Geometry.IntSize) or isinstance(obj, Geometry.IntRect) or isinstance(obj, Geometry.FloatPoint) or isinstance(obj, Geometry.FloatSize) or isinstance(obj, Geometry.FloatRect):
                        return tuple(obj)
                    else:
                        return json.JSONEncoder.default(self, obj)

            json_io = io.StringIO()
            json.dump(Utility.clean_dict(properties), json_io, cls=JSONEncoder)
            json_str = json_io.getvalue()

            self.__dataset.attrs["properties"] = json_str

    def __ensure_dataset(self):
        with self.__lock:
            self.__ensure_open()
            if self.__dataset is None:
                if "data" in self.__fp:
                    self.__dataset = self.__fp["data"]
                else:
                    self.__dataset = self.__fp.create_dataset("data", data=numpy.empty((0,)))

    def write_data(self, data, file_datetime):
        with self.__lock:
            assert data is not None
            self.__ensure_open()
            json_properties = None
            # handle three cases:
            #   1 - 'data' doesn't yet exist (require_dataset)
            #   2 - 'data' exists but is a different size (delete, then require_dataset)
            #   3 - 'data' exists and is the same size (overwrite)
            if not "data" in self.__fp:
                # case 1
                chunks = get_write_chunk_shape_for_data(data.shape, data.dtype)
                self.__dataset = self.__fp.require_dataset("data", shape=data.shape, dtype=data.dtype, chunks=chunks)
            else:
                if self.__dataset is None:
                    self.__dataset = self.__fp["data"]
                if self.__dataset.shape != data.shape or self.__dataset.dtype != data.dtype:
                    # case 2
                    json_properties = self.__dataset.attrs.get("properties", "")
                    self.__dataset = None
                    self.__fp.close()
                    self.__fp = None
                    os.remove(self.__file_path)
                    self.__ensure_open()
                    chunks = get_write_chunk_shape_for_data(data.shape, data.dtype)
                    self.__dataset = self.__fp.require_dataset("data", shape=data.shape, dtype=data.dtype, chunks=chunks)
            self.__copy_data(data)
            if json_properties is not None:
                self.__dataset.attrs["properties"] = json_properties
            self.__fp.flush()

    def reserve_data(self, data_shape: typing.Tuple[int, ...], data_dtype: numpy.dtype, file_datetime) -> None:
        # reserve data of the given shape and dtype, filled with zeros
        with self.__lock:
            self.__ensure_open()
            json_properties = None
            # first read existing properties and then close existing data set and file.
            if "data" in self.__fp:
                if self.__dataset is None:
                    self.__dataset = self.__fp["data"]
                json_properties = self.__dataset.attrs.get("properties", "")
                self.__dataset = None
                self.__fp.close()
                self.__fp = None
                os.remove(self.__file_path)
                self.__ensure_open()
            # reserve the data
            chunks = get_write_chunk_shape_for_data(data_shape, data_dtype)
            self.__dataset = self.__fp.require_dataset("data", shape=data_shape, dtype=data_dtype, fillvalue=0, chunks=chunks)
            if json_properties is not None:
                self.__dataset.attrs["properties"] = json_properties
            self.__fp.flush()

    def __copy_data(self, data):
        if id(data) != id(self.__dataset):
            self.__dataset[:] = data
            self._write_count += 1

    def write_properties(self, properties, file_datetime):
        with self.__lock:
            self.__ensure_open()
            self.__ensure_dataset()
            self.__write_properties_to_dataset(properties)
            self.__fp.flush()

    def read_properties(self):
        with self.__lock:
            self.__ensure_open()
            self.__ensure_dataset()
            json_properties = self.__dataset.attrs.get("properties", "")
            return json.loads(json_properties)

    def read_data(self):
        with self.__lock:
            self.__ensure_open()
            self.__ensure_dataset()
            if self.__dataset.shape == (0, ):
                return None
            return self.__dataset

    def remove(self):
        if self.__fp:
            self.__dataset = None
            self.__fp.close()
            self.__fp = None
        if os.path.isfile(self.__file_path):
            os.remove(self.__file_path)



# system imports
import ast
import collections
import contextlib
import copy
import functools
import gettext
import os
import threading
import traceback
import typing
import re
import locale
from importlib import reload
import sys
import json
import subprocess
import functools

# typing
from typing import AbstractSet

# third part imports
import numpy

# local libraries
from nion.swift.model import PlugInManager
from nion.swift.model import Utility
from nion.swift import FacadeQueued
from nion.swift import Panel
from nion.ui import Declarative
from nion.ui import Dialog
from nion.ui import Widgets
from nion.utils import Converter
from nion.utils import Selection
from nion.utils import Geometry

if typing.TYPE_CHECKING:
    from nion.swift import DocumentController

_ = gettext.gettext


def pose_get_string_message_box(ui, message_column, caption, text, accepted_fn, rejected_fn=None, accepted_text=None, rejected_text=None):
    if accepted_text is None:
        accepted_text = _("OK")
    if rejected_text is None:
        rejected_text = _("Cancel")
    message_box_widget = ui.create_column_widget()  # properties={"stylesheet": "background: #FFD"}
    caption_row = ui.create_row_widget()
    caption_row.add_spacing(12)
    caption_row.add(ui.create_label_widget(caption))
    caption_row.add_stretch()
    inside_row = ui.create_row_widget()

    def reject_button_clicked():
        if rejected_fn:
            rejected_fn()
        return False

    def accept_button_clicked():
        accepted_fn(string_edit_widget.text)
        return False

    string_edit_widget = ui.create_line_edit_widget()
    string_edit_widget.text = text
    string_edit_widget.on_return_pressed = accept_button_clicked
    string_edit_widget.on_escape_pressed = reject_button_clicked
    reject_button = ui.create_push_button_widget(rejected_text)
    reject_button.on_clicked = reject_button_clicked
    accepted_button = ui.create_push_button_widget(accepted_text)
    accepted_button.on_clicked = accept_button_clicked
    inside_row.add_spacing(12)
    inside_row.add(string_edit_widget)
    inside_row.add_spacing(12)
    inside_row.add(reject_button)
    inside_row.add_spacing(12)
    inside_row.add(accepted_button)
    inside_row.add_stretch()
    message_box_widget.add_spacing(6)
    message_box_widget.add(caption_row)
    message_box_widget.add_spacing(4)
    message_box_widget.add(inside_row)
    message_box_widget.add_spacing(4)
    message_column.add(message_box_widget)
    string_edit_widget.select_all()
    string_edit_widget.focused = True
    return message_box_widget


def pose_confirmation_message_box(ui, message_column, caption, accepted_fn, rejected_fn=None, accepted_text=None, rejected_text=None, display_rejected=True):
    if accepted_text is None:
        accepted_text = _("OK")
    if rejected_text is None:
        rejected_text = _("Cancel")
    message_box_widget = ui.create_column_widget()  # properties={"stylesheet": "background: #FFD"}

    def reject_button_clicked():
        if rejected_fn:
            rejected_fn()
        return False

    def accept_button_clicked():
        accepted_fn()
        return False

    reject_button = ui.create_push_button_widget(rejected_text)
    reject_button.on_clicked = reject_button_clicked
    accepted_button = ui.create_push_button_widget(accepted_text)
    accepted_button.on_clicked = accept_button_clicked
    caption_row = ui.create_row_widget()
    caption_row.add_spacing(12)
    caption_row.add(ui.create_label_widget(caption))
    if display_rejected:
        caption_row.add_spacing(12)
        caption_row.add(reject_button)
    caption_row.add_spacing(12)
    caption_row.add(accepted_button)
    caption_row.add_stretch()
    message_box_widget.add_spacing(6)
    message_box_widget.add(caption_row)
    message_box_widget.add_spacing(4)
    message_column.add(message_box_widget)
    return message_box_widget


class ScriptListItem:
    __version__ = "2"

    def __init__(self, full_path: str, indent: int = 0, indent_level: int = 0, show_dirname: bool = True):
        self.__full_path = os.path.abspath(full_path)
        self._exists = None
        self.indent = indent
        self.indent_level = indent_level
        self.show_dirname = show_dirname

    @property
    def full_path(self) -> str:
        return self.__full_path

    @property
    def basename(self) -> str:
        return os.path.basename(self.full_path)

    @property
    def dirname(self) -> str:
        return os.path.dirname(self.full_path)

    @property
    def exists(self) -> bool:
        return self._exists

    def check_existance(self) -> bool:
        self._exists = os.path.exists(self.full_path)
        return self._exists

    def to_dict(self) -> dict:
        return {"__version__": ScriptListItem.__version__,
                "__type__": "ScriptListItem",
                "full_path": self.full_path,
                "indent": self.indent,
                "indent_level": self.indent_level,
                "show_dirname": self.show_dirname,
                "_exists": self._exists}

    @classmethod
    def from_dict(cls, cls_dict: dict) -> "ScriptListItem":
        version = cls_dict.pop("__version__")
        if version != cls.__version__:
            raise ValueError(f"Cannot read the dictionary provided. Dictionary version '{version}' != implemented version '{cls.__version__}'.")
        instance = cls(cls_dict.pop("full_path"))
        for key, value in cls_dict.items():
            attr_name = key
            if key.startswith("__"):
                attr_name = "_" + cls.__name__ + key
            if hasattr(instance, attr_name):
                setattr(instance, attr_name, value)
        return instance

    def to_json_string(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json_string(cls, json_string: str) -> "ScriptListItem":
        return cls.from_dict(json.loads(json_string))

    # Used by "sort"
    def __lt__(self, other) -> bool:
        if isinstance(other, FolderListItem):
            return False
        if isinstance(other, ScriptListItem):
            return locale.strxfrm(self.basename) < locale.strxfrm(other.basename)
        return NotImplemented


class FolderListItem(ScriptListItem):

    def __init__(self, full_path: str, content: typing.Optional[typing.List[ScriptListItem]] = None, indent: int = 0, indent_level: int = 0, show_dirname: bool = True):
        super().__init__(full_path, indent=indent, indent_level=indent_level, show_dirname=show_dirname)
        self.__content = content if content is not None else list()
        self.folder_closed = True

    @property
    def content(self) -> typing.List[ScriptListItem]:
        return self.__content

    def update_content_from_file_system(self, filter_pattern: typing.Optional[str] = None):
        if os.path.isdir(self.full_path):
            dirlist = os.listdir(self.full_path)
            filtered_items = list()
            for item in dirlist:
                if filter_pattern is None or re.search(filter_pattern, item):
                    indent_level = self.indent_level + 1
                    filtered_items.append(ScriptListItem(os.path.join(self.full_path, item), indent=indent_level*20, indent_level=indent_level, show_dirname=False))
            self.__content = filtered_items

    def to_dict(self) -> dict:
        return {"__version__": FolderListItem.__version__,
                "__type__": "FolderListItem",
                "full_path": self.full_path,
                "indent": self.indent,
                "indent_level": self.indent_level,
                "show_dirname": self.show_dirname,
                "_exists": self._exists,
                "__content": [item.to_dict() for item in self.content],
                "folder_closed": self.folder_closed}

    @classmethod
    def from_dict(cls, cls_dict: dict) -> "FolderListItem":
        content = cls_dict.pop("__content", list())
        instance = super(FolderListItem, cls).from_dict(cls_dict)
        instance._FolderListItem__content = [ScriptListItem.from_dict(item) for item in content]
        return instance

    # Used by "sort"
    def __lt__(self, other) -> bool:
        if isinstance(other, FolderListItem):
            return locale.strxfrm(self.basename) < locale.strxfrm(other.basename)
        if isinstance(other, ScriptListItem):
            return True
        return NotImplemented


def _upgrade_item_dict_to_version_2(item_dict):
    if "__content" in item_dict:
        [_upgrade_item_dict_to_version_2(content_dict) for content_dict in item_dict["__content"]]
    if item_dict["__version__"] == "2":
        return item_dict
    if item_dict.get("indent") > 0:
        item_dict["indent_level"] = 1
    else:
        item_dict["indent_level"] = 0
    item_dict["__version__"] = "2"
    return item_dict


def _create_list_item_from_dict(item_dict) -> ScriptListItem:
    type_ = item_dict.pop("__type__")
    item_dict = _upgrade_item_dict_to_version_2(item_dict)
    return globals()[type_].from_dict(item_dict)


def _build_sorted_scripts_list(scripts_list: typing.List[ScriptListItem]) -> typing.List[ScriptListItem]:
        filtered_items = []
        for item in scripts_list:
            if item.indent_level == 0:
                filtered_items.append(item)
        filtered_items.sort()
        set_items = []
        for item in filtered_items:
            set_items.append(item)
            if isinstance(item, FolderListItem):
                if not item.folder_closed:
                    for content_item in item.content:
                        set_items.append(content_item)
        return set_items


def open_location(location: str) -> None:
    if sys.platform == "darwin":
        subprocess.Popen(["open", str(location)])
    elif sys.platform == 'win32':
        subprocess.run(['explorer', str(location)])
    elif sys.platform == 'linux':
        subprocess.check_call(['xdg-open', '--', str(location)])


class ScriptListCanvasItemDelegate(Widgets.ListCanvasItemDelegate):

    def __init__(self, ui, document_controller, update_list_fn: typing.Callable[[], typing.Any]):
        super().__init__()
        self.__ui = ui
        self.__document_controller = document_controller
        self.__update_list_fn = update_list_fn

    @staticmethod
    def _closed_folder_icon_strings() -> typing.Tuple[str, str]:
        if sys.platform == "win32":
            triangle = "\N{BLACK MEDIUM RIGHT-POINTING TRIANGLE CENTRED} "
        else:
            triangle = f"\N{BLACK RIGHT-POINTING TRIANGLE} "
        return triangle, f"\N{FILE FOLDER} "

    @staticmethod
    def _open_folder_icon_strings() -> typing.Tuple[str, str]:
        if sys.platform == "win32":
            triangle = "\N{BLACK MEDIUM DOWN-POINTING TRIANGLE CENTRED} "
        else:
            triangle = "\N{BLACK DOWN-POINTING TRIANGLE} "
        return triangle, f"\N{OPEN FILE FOLDER} "

    @staticmethod
    def _file_icon_string() -> str:
        return f"\N{PAGE FACING UP} "

    @staticmethod
    def _major_font_size() -> str:
        return "12px"

    @staticmethod
    def _minor_font_size() -> str:
        return "10px"

    @staticmethod
    def _major_font_color() -> str:
        return "#000"

    @staticmethod
    def _minor_font_color() -> str:
        return "#888"

    def close(self):
        ...

    def mouse_pressed_in_item(self, mouse_index: int, pos: Geometry.IntPoint, modifiers) -> bool:
        display_item = self.items[mouse_index]
        width = self.__ui.get_font_metrics(self._major_font_size(), self._closed_folder_icon_strings()[0]).width
        if isinstance(display_item, FolderListItem) and display_item.indent - 4 < pos.x < display_item.indent + width + 2:
            display_item.folder_closed = not display_item.folder_closed
            self.__update_list_fn()
            return True
        return False

    def context_menu_event(self, index: int, x: int, y: int, gx: int, gy: int) -> bool:
        display_item = self.items[index]
        menu = self.__document_controller.create_context_menu()
        if isinstance(display_item, FolderListItem):
            menu.add_menu_item(_("Open Folder"), functools.partial(open_location, display_item.full_path))
        elif isinstance(display_item, ScriptListItem):
            menu = self.__document_controller.create_context_menu()
            menu.add_menu_item(_("Open Containing Folder"), functools.partial(open_location, display_item.dirname))
        menu.popup(gx, gy)
        return True

    def __calculate_indent(self, display_item):
        # An item that can cause an indent_level > 0 is always an open folder
        triangle_string, icon_string = self._open_folder_icon_strings()
        return 4 + self.__ui.get_font_metrics(self._major_font_size(), triangle_string + icon_string).width * display_item.indent_level

    def paint_item(self, drawing_context, display_item, rect, is_selected):
        if isinstance(display_item, FolderListItem):
            if display_item.folder_closed:
                triangle_string, icon_string = self._closed_folder_icon_strings()
            else:
                triangle_string, icon_string = self._open_folder_icon_strings()
        else:
            triangle_string = str()
            icon_string = self._file_icon_string()

        icon_offset = self.__ui.get_font_metrics(self._major_font_size(), self._closed_folder_icon_strings()[0]).width
        icon_width = self.__ui.get_font_metrics(self._major_font_size(), icon_string).width

        if isinstance(display_item, ScriptListItem):
            with drawing_context.saver():

                drawing_context.fill_style = self._major_font_color()
                drawing_context.font = self._major_font_size()
                drawing_context.text_align = "left"
                drawing_context.text_baseline = "bottom"
                name_string = display_item.basename
                display_item.indent = self.__calculate_indent(display_item)
                drawing_context.fill_text(triangle_string, rect[0][1] + display_item.indent, rect[0][0] + 20 - 4)
                drawing_context.fill_text(icon_string + name_string, rect[0][1] + display_item.indent + icon_offset, rect[0][0] + 20 - 4)
                drawing_context.fill_style = self._minor_font_color()
                drawing_context.font = self._minor_font_color()
                name_width = self.__ui.get_font_metrics(self._major_font_size(), name_string).width
                if display_item.exists is not None and not display_item.exists:
                    type_str = "Folder" if type(display_item) is FolderListItem else "File"
                    drawing_context.fill_text(f"({type_str} not found)",
                                              rect[0][1] + 4 + display_item.indent + 4 + icon_offset + icon_width + name_width,
                                              rect[0][0] + 20 - 4)
                elif display_item.show_dirname:
                    drawing_context.fill_text(f"({display_item.dirname})",
                                              rect[0][1] + 4 + display_item.indent + 4 + icon_offset + icon_width + name_width,
                                              rect[0][0] + 20 - 4)


class RunScriptDialog(Dialog.ActionDialog):

    def __init__(self, document_controller: "DocumentController.DocumentController"):
        ui = document_controller.ui
        super().__init__(ui, _("Interactive Dialog"), parent_window=document_controller, persistent_id="ScriptsDialog")

        self.ui = ui
        self.document_controller = document_controller

        self.script_filter_pattern = "\.py$"

        self._create_menus()

        self.__cancelled = False

        self.__thread = None

        properties = dict()
        properties["min-height"] = 180
        properties["min-width"] = 540

        self.__output_widget = self.ui.create_text_edit_widget(properties)
        self.__output_widget.set_text_font(Panel.Panel.get_monospace_text_font())
        self.__output_widget.set_line_height_proportional(Panel.Panel.get_monospace_proportional_line_height())

        self.__message_column = ui.create_column_widget()

        self.__message_column.add(self.__make_cancel_row())

        # Load the list of known scripts. Upgrade from old storage system ("interactive_scripts_0")
        items = []
        items_str = self.ui.get_persistent_string("interactive_scripts_1", "")
        if items_str:
            for item_dict in json.loads(items_str):
                items.append(_create_list_item_from_dict(item_dict))
        else:
            items_old = self.ui.get_persistent_object("interactive_scripts_0", list())
            for item in items_old:
                items.append(ScriptListItem(item))
        if items:
            items = _build_sorted_scripts_list(items)
            self.ui.set_persistent_string("interactive_scripts_1", json.dumps([item.to_dict() for item in items]))

        self.__new_path_entries = []

        for item in items:
            if isinstance(item, FolderListItem):
                full_path = item.full_path
                self.__new_path_entries.append(full_path)
                if not full_path in sys.path:
                    sys.path.append(full_path)

        def selected_changed(indexes: AbstractSet[int]) -> None:
            run_button_widget.enabled = len(indexes) == 1

        def add_clicked() -> None:
            add_dir = self.ui.get_persistent_string("import_directory", "")
            file_paths, filter_str, directory = self.get_file_paths_dialog(_("Add Scripts"), add_dir, "Python Files (*.py)", "Python Files (*.py)")
            self.ui.set_persistent_string("import_directory", directory)
            items = self.scripts_list_widget.items
            items.extend([ScriptListItem(file_path) for file_path in file_paths])
            self.update_scripts_list(items)

        def add_folder_clicked() -> None:
            add_dir = self.ui.get_persistent_string("import_directory", "")
            existing_directory, directory = self.ui.get_existing_directory_dialog(_("Add Scripts Folder"), add_dir)
            if existing_directory:
                new_folder = FolderListItem(existing_directory)
                new_folder.update_content_from_file_system(filter_pattern=self.script_filter_pattern)
                full_path = new_folder.full_path
                if not full_path in sys.path:
                    sys.path.append(full_path)
                    self.__new_path_entries.append(full_path)
                items = self.scripts_list_widget.items
                items.append(new_folder)
                self.update_scripts_list(items)
            else:
                self.rebuild_scripts_list()

        def remove_clicked() -> None:
            indexes = list(self.scripts_list_widget.selected_items)
            new_items = []
            for i, item in enumerate(self.scripts_list_widget.items):
                if not i in indexes:
                    new_items.append(item)
            self.update_scripts_list(new_items)

        def run_clicked() -> None:
            indexes = self.scripts_list_widget.selected_items
            if len(indexes) == 1:
                script_item = self.scripts_list_widget.items[list(indexes)[0]]
                script_item.check_existance()
                # Use "type" instead of "isinstance" to exclude subclasses from matching
                if type(script_item) is ScriptListItem and script_item.exists:
                    script_path = script_item.full_path
                    self.run_script(script_path)

        def item_selected(index: int) -> bool:
            run_clicked()
            return True

        self.scripts_list_widget = Widgets.ListWidget(ui, ScriptListCanvasItemDelegate(ui, document_controller, self.rebuild_scripts_list), items=items, selection_style=Selection.Style.single_or_none, border_color="#888", properties={"min-height": 200, "min-width": 560, "size-policy-vertical": "expanding"})
        self.scripts_list_widget.on_selection_changed = selected_changed
        self.scripts_list_widget.on_item_selected = item_selected
        self.rebuild_scripts_list()

        add_button_widget = ui.create_push_button_widget(_("Add..."))
        add_button_widget.on_clicked = add_clicked

        add_folder_button_widget = ui.create_push_button_widget(_("Add Folder..."))
        add_folder_button_widget.on_clicked = add_folder_clicked

        remove_button_widget = ui.create_push_button_widget(_("Remove"))
        remove_button_widget.on_clicked = remove_clicked

        run_button_widget = ui.create_push_button_widget(_("Run"))
        run_button_widget.on_clicked = run_clicked
        run_button_widget.enabled = False

        list_widget_row = ui.create_row_widget()
        list_widget_row.add_spacing(8)
        list_widget_row.add(self.scripts_list_widget)
        list_widget_row.add_spacing(8)

        close_button_widget = ui.create_push_button_widget(_("Close"))
        close_button_widget.on_clicked = self.request_close

        button_row = ui.create_row_widget()
        button_row.add_spacing(12)
        button_row.add(add_button_widget)
        button_row.add_spacing(4)
        button_row.add(add_folder_button_widget)
        button_row.add_spacing(4)
        button_row.add(remove_button_widget)
        button_row.add_stretch()
        button_row.add(run_button_widget)
        button_row.add_spacing(12)

        select_column = ui.create_column_widget()
        select_column.add_spacing(8)
        select_column.add(list_widget_row)
        select_column.add_spacing(8)
        select_column.add(button_row)
        select_column.add_spacing(8)

        run_column = ui.create_column_widget()
        run_column.add(self.__output_widget)
        run_column.add_spacing(6)
        run_column.add(self.__message_column)

        self.__stack = ui.create_stack_widget()

        self.__stack.add(select_column)
        self.__stack.add(run_column)

        self.content.add(self.__stack)

        self.__sync_events = set()

        self.__lock = threading.RLock()

        self.__q = collections.deque()
        self.__output_queue = collections.deque()

        self.__is_closed = False

    def close(self):
        self.__is_closed = True
        for sync_event in self.__sync_events:
            sync_event.set()
        if self.__thread:
            self.__thread.join()
        self.document_controller.clear_task("ui_" + str(id(self)))
        self.document_controller.clear_task("run_" + str(id(self)))
        self.document_controller.clear_task("show_" + str(id(self)))
        self.document_controller.clear_task("print_" + str(id(self)))
        super().close()

    @property
    def cancelled(self):
        return self.__cancelled

    def update_scripts_list(self, new_scripts_list):
        for item in new_scripts_list:
            if isinstance(item, FolderListItem):
                item.update_content_from_file_system(filter_pattern=self.script_filter_pattern)
        items = _build_sorted_scripts_list(new_scripts_list)
        for item in items:
            item.check_existance()
        self.scripts_list_widget.items = items
        self.ui.set_persistent_string("interactive_scripts_1", json.dumps([item.to_dict() for item in self.scripts_list_widget.items]))

    def rebuild_scripts_list(self):
        self.update_scripts_list(self.scripts_list_widget.items)

    def __make_cancel_row(self):
        def cancel_script():
            self.__cancelled = True

        cancel_button = self.ui.create_push_button_widget(_("Cancel"))
        cancel_button.on_clicked = cancel_script
        cancel_row = self.ui.create_row_widget()
        cancel_row.add_stretch()
        cancel_row.add(cancel_button)
        cancel_row.add_spacing(12)
        return cancel_row

    def run_script(self, script_path: str) -> None:
        # Reload modules that are in on of the script folders.
        current_modules = list(sys.modules.values())
        for module in current_modules:
            for path in self.__new_path_entries:
                try:
                    do_reload = os.path.abspath(module.__file__)[:len(path)] == path
                except (AttributeError, ValueError, TypeError):
                    pass
                else:
                    try:
                        if do_reload:
                            reload(module)
                    except:
                        self.print(traceback.format_exc())
                        self.__stack.current_index = 1
                        self.continue_after_parse_error(script_path)
                        return

        self.__output_widget.text = None

        script_name = os.path.basename(script_path)

        self.title = os.path.splitext(script_name)[0]

        try:
            with open(script_path) as f:
                script = f.read()

            script_ast = ast.parse(script, script_name, 'exec')
        except Exception as e:
            self.print(str(e))
            self.__stack.current_index = 1
            self.continue_after_parse_error(script_path)
            return

        class AddCallFunctionNodeTransformer(ast.NodeTransformer):
            def __init__(self, func_id, arg_id):
                self.__func_id = func_id
                self.__arg_id = arg_id

            def visit_Module(self, node):
                name_expr = ast.Name(id=self.__func_id, ctx=ast.Load())
                arg_expr = ast.Name(id=self.__arg_id, ctx=ast.Load())
                call_expr = ast.Expr(value=ast.Call(func=name_expr, args=[arg_expr], keywords=[]))
                new_node = copy.deepcopy(node)
                new_node.body.append(call_expr)
                ast.fix_missing_locations(new_node)
                return new_node

        # if script_main exists, add a node to call it
        for node in script_ast.body:
            if getattr(node, "name", None) == "script_main":
                script_ast = AddCallFunctionNodeTransformer('script_main', 'api_broker').visit(script_ast)

        compiled = compile(script_ast, script_name, 'exec')

        def run_it(compiled, interactive_session):
            class APIBroker:
                def get_interactive(self, version=None):
                    actual_version = "1.0.0"
                    if Utility.compare_versions(version, actual_version) > 0:
                        raise NotImplementedError("API requested version %s is greater than %s." % (version, actual_version))
                    return interactive_session
                def get_api(self, version, ui_version=None):
                    ui_version = ui_version if ui_version else "~1.0"
                    api = PlugInManager.api_broker_fn(version, ui_version)
                    queued_api = FacadeQueued.API(api, None)
                    queued_api._queue_task = api.queue_task
                    return queued_api
                def get_ui(self, version):
                    actual_version = "1.0.0"
                    if Utility.compare_versions(version, actual_version) > 0:
                        raise NotImplementedError("API requested version %s is greater than %s." % (version, actual_version))
                    return Declarative.DeclarativeUI()
            try:
                g = dict()
                g["api_broker"] = APIBroker()
                g["print"] = self.print
                g["input"] = self.get_string

                print_fn = self.print
                self.__cancelled = False # Reset cancelled flag to make "Run again" work after a script was cancelled
                class StdoutCatcher:
                    def __init__(self):
                        pass
                    def write(self, stuff):
                        print_fn(stuff.rstrip())
                    def flush(self):
                        pass
                stdout = StdoutCatcher()
                with contextlib.redirect_stdout(stdout), contextlib.redirect_stderr(stdout):
                    exec(compiled, g)
            except Exception:
                self.print("{}: {}".format(_("Error"), traceback.format_exc()))

        self.__stack.current_index = 1

        self.run(script_path, functools.partial(run_it, compiled))

    def continue_after_parse_error(self, script_path):
        def func_continue():
            result = self.confirm(_("Finished"), _("Run Again"), _("Close"))

            with self.__lock:
                if result:
                    self.__q.append(functools.partial(self.run_script, script_path))
                else:
                    self.__q.append(self.request_close)
                self.document_controller.add_task("run_" + str(id(self)), self.__handle_output_and_q)

        self.__thread = threading.Thread(target=func_continue)
        self.__thread.start()

    def run(self, script_path, func):
        def func_run(func):
            try:
                func(self)
            except Exception:
                pass

            if not self.__is_closed:
                result = self.confirm(_("Finished"), _("Run Again"), _("Close"))

                with self.__lock:
                    if result:
                        self.__q.append(functools.partial(self.run_script, script_path))
                    else:
                        self.__q.append(self.request_close)
                    self.document_controller.add_task("run_" + str(id(self)), self.__handle_output_and_q)

        self.__thread = threading.Thread(target=func_run, args=(func,))
        self.__thread.start()

    def __handle_output_and_q(self):
        func = None
        with self.__lock:
            while len(self.__output_queue) > 0:
                self.__output_widget.move_cursor_position("end")
                self.__output_widget.append_text(self.__output_queue.popleft())
            if len(self.__q) > 0:
                func = self.__q.popleft()
        if callable(func):
            func()

    def print(self, text):
        with self.__lock:
            self.__output_queue.append(str(text))
            self.document_controller.add_task("print_" + str(id(self)), self.__handle_output_and_q)

    def print_debug(self, text):
        self.print(text)

    def print_info(self, text):
        self.print(text)

    def print_warn(self, text):
        self.print(text)

    def get_string(self, prompt, default_str=None) -> typing.Optional[str]:
        """Return a string value that the user enters. Raises exception for cancel."""
        with self.sync_event() as accept_event:
            result = None

            def perform():
                def accepted(text):
                    nonlocal result
                    result = text
                    accept_event.set()

                def rejected():
                    accept_event.set()

                self.__message_column.remove_all()
                pose_get_string_message_box(self.ui, self.__message_column, prompt, str(default_str), accepted, rejected)

            with self.__lock:
                self.__q.append(perform)
                self.document_controller.add_task("ui_" + str(id(self)), self.__handle_output_and_q)
            accept_event.wait()
            if self.__is_closed:
                raise Exception("Cancel")

        def update_message_column():
            self.__message_column.remove_all()
            self.__message_column.add(self.__make_cancel_row())
        self.document_controller.add_task("ui_" + str(id(self)), update_message_column)
        if result is None:
            raise Exception("Cancel")
        return result

    def get_integer(self, prompt: str, default_value: int=0) -> int:
        converter = Converter.IntegerToStringConverter()
        result = self.get_string(prompt, converter.convert(default_value))
        return converter.convert_back(result)

    def get_float(self, prompt: str, default_value: float=0, format_str: str=None) -> float:
        converter = Converter.FloatToStringConverter(format_str)
        result = self.get_string(prompt, converter.convert(default_value))
        return converter.convert_back(result)

    def show_ndarray(self, data: numpy.ndarray, title:str = None) -> None:
        with self.sync_event() as accept_event:

            def perform():
                result_display_panel = self.document_controller.next_result_display_panel()
                if result_display_panel:
                    data_item = self.document_controller.add_data(data, title)
                    display_item = self.document_controller.document_model.get_display_item_for_data_item(data_item)
                    result_display_panel.set_display_panel_display_item(display_item)
                    result_display_panel.request_focus()
                accept_event.set()

            with self.__lock:
                self.__q.append(perform)
                self.document_controller.add_task("show_" + str(id(self)), self.__handle_output_and_q)
            accept_event.wait()
            if self.__is_closed:
                raise Exception("Cancel")

    def __register_sync_event(self, sync_event: threading.Event) -> None:
        self.__sync_events.add(sync_event)

    def __unregister_sync_event(self, sync_event: threading.Event) -> None:
        self.__sync_events.remove(sync_event)

    @contextlib.contextmanager
    def sync_event(self):
        sync_event = threading.Event()
        self.__register_sync_event(sync_event)
        yield sync_event
        self.__unregister_sync_event(sync_event)

    def __accept_reject(self, prompt, accepted_text, rejected_text, display_rejected):
        """Return a boolean value for accept/reject."""
        with self.sync_event() as accept_event:
            result = False

            def perform():
                def accepted():
                    nonlocal result
                    result = True
                    accept_event.set()

                def rejected():
                    nonlocal result
                    result = False
                    accept_event.set()

                self.__message_column.remove_all()
                pose_confirmation_message_box(self.ui, self.__message_column, prompt, accepted, rejected, accepted_text, rejected_text, display_rejected)
                #self.__message_column.add(self.__make_cancel_row())

            with self.__lock:
                self.__q.append(perform)
                self.document_controller.add_task("ui_" + str(id(self)), self.__handle_output_and_q)
            accept_event.wait()
            if self.__is_closed:
                raise Exception("Cancel")

        def update_message_column():
            self.__message_column.remove_all()
            self.__message_column.add(self.__make_cancel_row())
        self.document_controller.add_task("ui_" + str(id(self)), update_message_column)
        return result

    def confirm_ok_cancel(self, prompt: str) -> bool:
        return self.__accept_reject(prompt, _("OK"), _("Cancel"), True)

    def confirm_yes_no(self, prompt: str) -> bool:
        return self.__accept_reject(prompt, _("Yes"), _("No"), True)

    def confirm(self, prompt: str, accepted_text: str, rejected_text: str) -> bool:
        return self.__accept_reject(prompt, accepted_text, rejected_text, True)

    def alert(self, prompt: str, button_label: str = None) -> None:
        self.__accept_reject(prompt, button_label, None, False)



# standard libraries
import logging
import unittest
import unittest.mock

# third party libraries
import numpy

# local libraries
from nion.swift import Application
from nion.swift import Facade
from nion.swift.model import DataItem
from nion.swift.test import TestContext
from nion.ui import TestUI


Facade.initialize()


def create_memory_profile_context() -> TestContext.MemoryProfileContext:
    return TestContext.MemoryProfileContext()


def press_ok_and_complete(*args, **kwargs):
    kwargs["completion_fn"]()

def press_ok_cancel_and_ok_complete(*args, **kwargs):
    kwargs["completion_fn"](True)

def press_ok_cancel_and_cancel_complete(*args, **kwargs):
    kwargs["completion_fn"](False)


class TestProfileClass(unittest.TestCase):

    def setUp(self):
        self.app = Application.Application(TestUI.UserInterface(), set_global=False)

    def tearDown(self):
        pass

    def test_closing_window_unloads_project_referemnce(self):
        with create_memory_profile_context() as profile_context:
            # use lower level calls to create the profile and open the window via the app
            profile = profile_context.create_profile()
            profile.read_profile()
            self.app._set_profile_for_test(profile)
            TestContext.add_project_memory(profile, load=False)
            document_controller = self.app.open_project_window(profile.project_references[0])
            # check conditions
            self.assertEqual(2, len(profile.project_references))
            self.assertEqual("loaded", profile.project_references[0].project_state)
            self.assertEqual("unloaded", profile.project_references[1].project_state)
            document_controller.request_close()
            self.assertEqual("unloaded", profile.project_references[0].project_state)
            self.assertEqual("unloaded", profile.project_references[1].project_state)

    def test_switching_projects_and_back(self):
        with create_memory_profile_context() as profile_context:
            # use lower level calls to create the profile and open the window via the app
            profile = profile_context.create_profile()
            profile.read_profile()
            self.app._set_profile_for_test(profile)
            TestContext.add_project_memory(profile, load=False)
            # load first project and check conditions
            document_controller = self.app.open_project_window(profile.project_references[0])
            self.assertEqual("loaded", profile.project_references[0].project_state)
            self.assertEqual("unloaded", profile.project_references[1].project_state)
            # switch projects and check conditions
            document_controller.request_close()
            document_controller = self.app.open_project_window(profile.project_references[1])
            self.assertEqual("unloaded", profile.project_references[0].project_state)
            self.assertEqual("loaded", profile.project_references[1].project_state)
            # switch projects again and check conditions
            document_controller.request_close()
            document_controller = self.app.open_project_window(profile.project_references[0])
            self.assertEqual("loaded", profile.project_references[0].project_state)
            self.assertEqual("unloaded", profile.project_references[1].project_state)
            document_controller.request_close()

    def test_switching_projects_with_related_items(self):
        # this tests that project gets removed from persistent object context
        with create_memory_profile_context() as profile_context:
            # use lower level calls to create the profile and open the window via the app
            profile = profile_context.create_profile()
            profile.read_profile()
            self.app._set_profile_for_test(profile)
            TestContext.add_project_memory(profile, load=False)
            # load first project and check conditions
            document_controller = self.app.open_project_window(profile.project_references[0])
            document_controller.document_model.append_data_item(DataItem.DataItem(numpy.zeros((2, 2))))
            document_controller._perform_processing(document_controller.document_model.display_items[0], document_controller.document_model.display_items[0].data_item, None, document_controller.document_model.get_fft_new)
            self.assertEqual("loaded", profile.project_references[0].project_state)
            self.assertEqual("unloaded", profile.project_references[1].project_state)
            self.assertEqual(2, len(document_controller.document_model.data_items))
            # switch projects and check conditions
            document_controller.request_close()
            document_controller = self.app.open_project_window(profile.project_references[1])
            self.assertEqual("unloaded", profile.project_references[0].project_state)
            self.assertEqual("loaded", profile.project_references[1].project_state)
            # switch projects again and check conditions
            document_controller.request_close()
            document_controller = self.app.open_project_window(profile.project_references[0])
            self.assertEqual("loaded", profile.project_references[0].project_state)
            self.assertEqual("unloaded", profile.project_references[1].project_state)
            self.assertEqual(2, len(document_controller.document_model.data_items))
            document_controller.request_close()

    def test_forget_project(self):
        with create_memory_profile_context() as profile_context:
            # use lower level calls to create the profile and open the window via the app
            profile = profile_context.create_profile()
            profile.read_profile()
            self.app._set_profile_for_test(profile)
            TestContext.add_project_memory(profile, load=False)
            # load first project and check conditions
            document_controller = self.app.open_project_window(profile.project_references[0])
            self.assertEqual("loaded", profile.project_references[0].project_state)
            self.assertEqual("unloaded", profile.project_references[1].project_state)
            # forget project and check conditions
            profile.remove_project_reference(profile.project_references[1])
            self.assertEqual(1, len(profile.project_references))
            self.assertEqual("loaded", profile.project_references[0].project_state)
            # clean up
            document_controller.close()

    def test_create_project(self):
        with create_memory_profile_context() as profile_context:
            # use lower level calls to create the profile and open the window via the app
            profile = profile_context.create_profile()
            profile.read_profile()
            self.app._set_profile_for_test(profile)
            document_controller = self.app.open_project_window(profile.project_references[0])
            # check preconditions
            self.assertEqual(1, len(self.app.windows))
            self.assertEqual(1, len(profile.project_references))
            self.assertEqual("loaded", profile.project_references[0].project_state)
            # add project, check
            project_reference2 = profile.add_project_memory(load=False)
            self.assertEqual(2, len(profile.project_references))
            self.assertEqual("loaded", profile.project_references[0].project_state)
            self.assertEqual("unloaded", profile.project_references[1].project_state)
            self.assertEqual(project_reference2, profile.project_references[1])
            self.assertEqual(1, len(self.app.windows))
            # clean up
            document_controller.close()

    def test_forget_loaded_project_not_allowed(self):
        with create_memory_profile_context() as profile_context:
            # use lower level calls to create the profile and open the window via the app
            profile = profile_context.create_profile()
            profile.read_profile()
            self.app._set_profile_for_test(profile)
            TestContext.add_project_memory(profile, load=False)
            # load first project and check conditions
            document_controller = self.app.open_project_window(profile.project_references[0])
            self.assertEqual("loaded", profile.project_references[0].project_state)
            self.assertEqual("unloaded", profile.project_references[1].project_state)
            # forget project and check conditions
            with self.assertRaises(Exception):
                profile.remove_project_reference(profile.project_references[0])
            self.assertEqual(2, len(profile.project_references))
            self.assertEqual("loaded", profile.project_references[0].project_state)
            self.assertEqual("unloaded", profile.project_references[1].project_state)
            # clean up
            document_controller.close()

    def test_launch_with_no_project_then_cancel_choose_dialog(self):
        with create_memory_profile_context() as profile_context:
            # use lower level calls to create the profile and open the window via the app
            profile = profile_context.create_profile(add_project=False)
            profile.read_profile()
            app = Application.Application(TestUI.UserInterface(), set_global=False)
            app._set_profile_for_test(profile)
            app.initialize(load_plug_ins=False)
            try:
                # ensure no project references
                self.assertEqual(0, len(profile.project_references))
                # set up mock calls to get through start call and check conditions
                app.show_choose_project_dialog = unittest.mock.Mock()
                # start the app
                app.start(profile=profile)
                # check the mock calls
                app.show_choose_project_dialog.assert_called_once()
                # ensure no project is loaded
                self.assertEqual(0, len(profile.project_references))
                self.assertEqual(0, len(app.windows))
            finally:
                app._set_profile_for_test(None)
                app.exit()
                app.deinitialize()

    def test_launch_with_no_project_then_cancel_open_dialog(self):
        with create_memory_profile_context() as profile_context:
            # use lower level calls to create the profile and open the window via the app
            profile = profile_context.create_profile(add_project=False)
            profile.read_profile()
            app = Application.Application(TestUI.UserInterface(), set_global=False)
            app._set_profile_for_test(profile)
            app.initialize(load_plug_ins=False)
            try:
                # ensure no project references
                self.assertEqual(0, len(profile.project_references))
                # set up mock calls to get through start call and check conditions
                app.ui.get_file_paths_dialog = unittest.mock.Mock(return_value=([], str(), str()))
                app.show_ok_dialog = unittest.mock.Mock()
                app.show_ok_dialog.side_effect = press_ok_and_complete
                app.show_choose_project_dialog = unittest.mock.Mock()
                app.show_choose_project_dialog.side_effect = app.show_open_project_dialog
                # start the app
                app.start(profile=profile)
                # check the mock calls
                app.ui.get_file_paths_dialog.assert_called_once()
                app.show_choose_project_dialog.assert_called_once()
                app.show_ok_dialog.assert_not_called()
                # ensure no project is loaded
                self.assertEqual(0, len(profile.project_references))
                self.assertEqual(0, len(app.windows))
            finally:
                app._set_profile_for_test(None)
                app.exit()
                app.deinitialize()

    def test_launch_with_no_project_then_open_a_project(self):
        with create_memory_profile_context() as profile_context:
            # use lower level calls to create the profile and open the window via the app
            profile = profile_context.create_profile(add_project=False)
            profile.read_profile()
            app = Application.Application(TestUI.UserInterface(), set_global=False)
            app._set_profile_for_test(profile)
            app.initialize(load_plug_ins=False)
            try:
                # ensure no project references
                self.assertEqual(0, len(profile.project_references))

                # set up mock calls to get through start call
                profile.open_project = unittest.mock.Mock()
                profile.open_project.side_effect = lambda x: TestContext.add_project_memory(profile, load=False)
                app.ui.get_file_paths_dialog = unittest.mock.Mock(return_value=(["PATH"], str(), str()))
                app.show_ok_dialog = unittest.mock.Mock()
                app.show_ok_dialog.side_effect = press_ok_and_complete
                app.show_choose_project_dialog = unittest.mock.Mock()
                app.show_choose_project_dialog.side_effect = app.show_open_project_dialog
                logging.getLogger("loader").setLevel = unittest.mock.Mock()  # ignore this call

                # start the app
                app.start(profile=profile)

                # check the mock calls
                app.ui.get_file_paths_dialog.assert_called_once()
                profile.open_project.assert_called_once()
                app.show_choose_project_dialog.assert_called_once()
                app.show_ok_dialog.assert_not_called()

                # ensure a single project is loaded
                self.assertEqual(1, len(profile.project_references))
                self.assertEqual("loaded", profile.project_references[0].project_state)
                self.assertEqual(1, len(app.windows))
            finally:
                app._set_profile_for_test(None)
                app.exit()
                app.deinitialize()

    def test_launch_with_no_project_then_open_a_project_with_error(self):
        # test both json error and general storage error
        for uuid_error, storage_error in ((True, False), (False, True)):
            with self.subTest(uuid_error=uuid_error, storage_error=storage_error):
                with create_memory_profile_context() as profile_context:
                    # use lower level calls to create the profile and open the window via the app
                    profile = profile_context.create_profile(add_project=False)
                    profile.read_profile()
                    app = Application.Application(TestUI.UserInterface(), set_global=False)
                    app._set_profile_for_test(profile)
                    app.initialize(load_plug_ins=False)
                    try:
                        # ensure no project references
                        self.assertEqual(0, len(profile.project_references))

                        # set up mock calls to get through start call
                        def setup_bad_project(args):
                            project_reference = TestContext.add_project_memory(profile, load=False, make_uuid_error=uuid_error, make_storage_error=storage_error)
                            return project_reference

                        app.show_choose_project_dialog = unittest.mock.Mock()
                        app.show_choose_project_dialog.side_effect = app.show_open_project_dialog
                        profile.open_project = unittest.mock.Mock()
                        profile.open_project.side_effect = setup_bad_project
                        app.ui.get_file_paths_dialog = unittest.mock.Mock()
                        app.ui.get_file_paths_dialog.side_effect = [(["PATH"], str(), str()), ([], str(), str())]
                        app.show_ok_dialog = unittest.mock.Mock()
                        app.show_ok_dialog.side_effect = press_ok_and_complete
                        logging.getLogger("loader").setLevel = unittest.mock.Mock()  # ignore this call

                        # start the app
                        app.start(profile=profile)

                        # check the mock calls
                        self.assertEqual(2, len(app.ui.get_file_paths_dialog.mock_calls))
                        profile.open_project.assert_called_once()
                        app.show_ok_dialog.assert_called_once()

                        # ensure a single project is loaded
                        self.assertEqual(1, len(profile.project_references))
                        self.assertEqual("invalid", profile.project_references[0].project_state)
                        self.assertEqual(0, len(app.windows))
                    finally:
                        app._set_profile_for_test(None)
                        app.exit()
                        app.deinitialize()

    def test_launch_with_no_project_then_open_a_project_needing_upgrade_and_cancel(self):
        with create_memory_profile_context() as profile_context:
            # use lower level calls to create the profile and open the window via the app
            profile = profile_context.create_profile(add_project=False)
            profile.read_profile()
            app = Application.Application(TestUI.UserInterface(), set_global=False)
            app._set_profile_for_test(profile)
            app.initialize(load_plug_ins=False)
            try:
                # ensure no project references
                self.assertEqual(0, len(profile.project_references))

                # set up mock calls to get through start call
                def setup_needs_upgrade_project(args):
                    project_reference = TestContext.add_project_memory(profile, load=False, d={"version": 0})
                    return project_reference

                profile.open_project = unittest.mock.Mock()
                profile.open_project.side_effect = setup_needs_upgrade_project
                app.ui.get_file_paths_dialog = unittest.mock.Mock()
                app.ui.get_file_paths_dialog = unittest.mock.Mock(return_value=(["PATH"], str(), str()))
                app.show_ok_cancel_dialog = unittest.mock.Mock()
                app.show_ok_cancel_dialog.side_effect = press_ok_cancel_and_cancel_complete
                app.show_choose_project_dialog = unittest.mock.Mock()
                app.show_choose_project_dialog.side_effect = app.show_open_project_dialog
                logging.getLogger("loader").setLevel = unittest.mock.Mock()  # ignore this call

                # start the app
                app.start(profile=profile)

                # check the mock calls
                self.assertEqual(1, len(app.ui.get_file_paths_dialog.mock_calls))
                profile.open_project.assert_called_once()
                app.show_choose_project_dialog.assert_called_once()
                app.show_ok_cancel_dialog.assert_called_once()

                # ensure a single project is loaded
                self.assertEqual(1, len(profile.project_references))
                self.assertEqual("needs_upgrade", profile.project_references[0].project_state)
                self.assertEqual(0, len(app.windows))
            finally:
                app._set_profile_for_test(None)
                app.exit()
                app.deinitialize()

    def test_launch_with_no_project_then_open_a_project_needing_upgrade_and_proceed(self):
        with create_memory_profile_context() as profile_context:
            # use lower level calls to create the profile and open the window via the app
            profile = profile_context.create_profile(add_project=False)
            profile.read_profile()
            app = Application.Application(TestUI.UserInterface(), set_global=False)
            app._set_profile_for_test(profile)
            app.initialize(load_plug_ins=False)
            try:
                # ensure no project references
                self.assertEqual(0, len(profile.project_references))

                # set up mock calls to get through start call
                def setup_needs_upgrade_project(args):
                    project_reference = TestContext.add_project_memory(profile, load=False, d={"version": 0})
                    return project_reference

                profile.open_project = unittest.mock.Mock()
                profile.open_project.side_effect = setup_needs_upgrade_project
                app.ui.get_file_paths_dialog = unittest.mock.Mock()
                app.ui.get_file_paths_dialog = unittest.mock.Mock(return_value=(["PATH"], str(), str()))
                app.show_ok_cancel_dialog = unittest.mock.Mock()
                app.show_ok_cancel_dialog.side_effect = press_ok_cancel_and_ok_complete
                app.show_choose_project_dialog = unittest.mock.Mock()
                app.show_choose_project_dialog.side_effect = app.show_open_project_dialog
                logging.getLogger("loader").setLevel = unittest.mock.Mock()  # ignore this call

                # start the app
                app.start(profile=profile)

                # check the mock calls
                self.assertEqual(1, len(app.ui.get_file_paths_dialog.mock_calls))
                profile.open_project.assert_called_once()
                app.show_choose_project_dialog.assert_called_once()
                app.show_ok_cancel_dialog.assert_called_once()

                # ensure a single project is loaded
                self.assertEqual(1, len(profile.project_references))
                self.assertEqual("loaded", profile.project_references[0].project_state)
                self.assertEqual(1, len(app.windows))
            finally:
                app._set_profile_for_test(None)
                app.exit()
                app.deinitialize()

    def test_launch_with_no_project_then_open_a_project_needing_upgrade_but_unable_to_upgrade(self):
        with create_memory_profile_context() as profile_context:
            # use lower level calls to create the profile and open the window via the app
            profile = profile_context.create_profile(add_project=False)
            profile.read_profile()
            app = Application.Application(TestUI.UserInterface(), set_global=False)
            app._set_profile_for_test(profile)
            app.initialize(load_plug_ins=False)
            try:
                # ensure no project references
                self.assertEqual(0, len(profile.project_references))

                # set up mock calls to get through start call
                def setup_needs_upgrade_project(args):
                    project_reference = TestContext.add_project_memory(profile, load=False, d={"version": 0})
                    return project_reference

                profile.open_project = unittest.mock.Mock()
                profile.open_project.side_effect = setup_needs_upgrade_project
                app.ui.get_file_paths_dialog = unittest.mock.Mock()
                app.ui.get_file_paths_dialog = unittest.mock.Mock(return_value=(["PATH"], str(), str()))
                app.show_choose_project_dialog = unittest.mock.Mock()
                app.show_choose_project_dialog.side_effect = app.show_open_project_dialog
                app.show_ok_cancel_dialog = unittest.mock.Mock()
                app.show_ok_cancel_dialog.side_effect = press_ok_cancel_and_ok_complete
                app.show_ok_dialog = unittest.mock.Mock()
                profile.upgrade = unittest.mock.Mock()
                profile.upgrade.side_effect = FileExistsError()
                logging.getLogger("loader").setLevel = unittest.mock.Mock()  # ignore this call

                # start the app
                app.start(profile=profile)

                # check the mock calls
                self.assertEqual(1, len(app.ui.get_file_paths_dialog.mock_calls))
                profile.open_project.assert_called_once()
                app.show_choose_project_dialog.assert_called_once()
                app.show_ok_cancel_dialog.assert_called_once()
                app.show_ok_dialog.assert_called_once()

                # ensure a single project is loaded
                self.assertEqual(1, len(profile.project_references))
                self.assertEqual(0, len(app.windows))
            finally:
                app._set_profile_for_test(None)
                app.exit()
                app.deinitialize()

    def test_switching_project_reloads_new_project_on_relaunch(self):
        with create_memory_profile_context() as profile_context:
            # use lower level calls to create the profile and open the window via the app
            profile = profile_context.create_profile(add_project=False)
            profile.read_profile()
            app = Application.Application(TestUI.UserInterface(), set_global=False)
            app._set_profile_for_test(profile)
            TestContext.add_project_memory(profile, load=False)
            TestContext.add_project_memory(profile, load=False)
            app.initialize(load_plug_ins=False)
            try:
                # ensure one project reference
                self.assertEqual(2, len(profile.project_references))
                profile.last_project_reference = profile.project_references[0].uuid
                logging.getLogger("loader").setLevel = unittest.mock.Mock()  # ignore this call

                # start the app
                app.start(profile=profile)

                # ensure a single project is loaded
                self.assertEqual(2, len(profile.project_references))
                self.assertEqual("loaded", profile.project_references[0].project_state)
                self.assertEqual("unloaded", profile.project_references[1].project_state)
                self.assertEqual(1, len(app.windows))
                self.assertEqual(profile.last_project_reference, profile.project_references[0].uuid)

                # switch project and check
                app.switch_project_reference(profile.project_references[1])
                self.assertEqual("unloaded", profile.project_references[0].project_state)
                self.assertEqual("loaded", profile.project_references[1].project_state)
                self.assertEqual(1, len(app.windows))
                self.assertEqual(profile.last_project_reference, profile.project_references[1].uuid)
            finally:
                app._set_profile_for_test(None)
                app.exit()
                app.deinitialize()

            # reload and check
            profile = profile_context.create_profile(add_project=False)
            profile.read_profile()
            app = Application.Application(TestUI.UserInterface(), set_global=False)
            app._set_profile_for_test(profile)
            app.initialize(load_plug_ins=False)
            try:
                # ensure two project references
                self.assertEqual(2, len(profile.project_references))
                profile.last_project_reference = profile.project_references[0].uuid

                # start the app
                app.start(profile=profile)

                # switch project and check
                app.switch_project_reference(profile.project_references[1])
                self.assertEqual("unloaded", profile.project_references[0].project_state)
                self.assertEqual("loaded", profile.project_references[1].project_state)
                self.assertEqual(1, len(app.windows))
                self.assertEqual(profile.last_project_reference, profile.project_references[1].uuid)
            finally:
                app._set_profile_for_test(None)
                app.exit()
                app.deinitialize()

    def test_switch_to_project_with_error(self):
        # test both json error and general storage error
        for uuid_error, storage_error in ((True, False), (False, True)):
            with self.subTest(uuid_error=uuid_error, storage_error=storage_error):
                with create_memory_profile_context() as profile_context:
                    # use lower level calls to create the profile and open the window via the app
                    profile = profile_context.create_profile(add_project=False)
                    profile.read_profile()
                    app = Application.Application(TestUI.UserInterface(), set_global=False)
                    app._set_profile_for_test(profile)
                    TestContext.add_project_memory(profile, load=False)
                    TestContext.add_project_memory(profile, load=False)
                    app.initialize(load_plug_ins=False)
                    try:
                        # ensure two project references
                        self.assertEqual(2, len(profile.project_references))
                        profile.last_project_reference = profile.project_references[0].uuid
                        logging.getLogger("loader").setLevel = unittest.mock.Mock()  # ignore this call
                        app.show_choose_project_dialog = unittest.mock.Mock()
                        app.show_choose_project_dialog.side_effect = app.show_open_project_dialog

                        # start the app
                        app.start(profile=profile)

                        # ensure a single project is loaded
                        self.assertEqual(2, len(profile.project_references))
                        self.assertEqual("loaded", profile.project_references[0].project_state)
                        self.assertEqual("unloaded", profile.project_references[1].project_state)
                        self.assertEqual(1, len(app.windows))
                        self.assertEqual(profile.last_project_reference, profile.project_references[0].uuid)

                        # switching to an error project should display an ok message with an error, then the file
                        # open dialog (which will be cancelled). no windows should be open at the end.

                        # set up mock calls to get through the switch.
                        if uuid_error:
                            profile.read_project = unittest.mock.Mock()
                            profile.read_project.side_effect = Exception()
                        if storage_error:
                            profile.project_references[1].make_storage = unittest.mock.Mock()
                            profile.project_references[1].make_storage.side_effect = Exception()
                        app.ui.get_file_paths_dialog = unittest.mock.Mock()
                        app.ui.get_file_paths_dialog.side_effect = [([], str(), str()), ([], str(), str())]
                        app.show_ok_dialog = unittest.mock.Mock()
                        app.show_ok_dialog.side_effect = press_ok_and_complete
                        logging.getLogger("loader").setLevel = unittest.mock.Mock()  # ignore this call

                        # switch project and check
                        app.switch_project_reference(profile.project_references[1])
                        app.ui.get_file_paths_dialog.assert_called_once()
                        app.show_ok_dialog.assert_called_once()
                        self.assertEqual("unloaded", profile.project_references[0].project_state)
                        self.assertEqual("unloaded", profile.project_references[1].project_state)
                        self.assertEqual(0, len(app.windows))
                        self.assertEqual(profile.last_project_reference, profile.project_references[0].uuid)
                    finally:
                        app._set_profile_for_test(None)
                        app.exit()
                        app.deinitialize()

    # TODO: creating new project
    # TODO: opening project from file
    # TODO: opening same project is not allowed
    # TODO: recent menu is sorted by date and limited
    # TODO: upgrade project



# standard libraries
import contextlib
import unittest

# third party libraries
import numpy

# local libraries
from nion.data import DataAndMetadata
from nion.swift import Application
from nion.swift.model import DataItem
from nion.swift.model import DisplayItem
from nion.swift.model import Graphics
from nion.swift.test import TestContext
from nion.ui import TestUI


class TestInfoPanelClass(unittest.TestCase):

    def setUp(self):
        self.app = Application.Application(TestUI.UserInterface(), set_global=False)

    def tearDown(self):
        pass

    def test_cursor_over_1d_composite_image(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((32,)))
            data_item2 = DataItem.DataItem(numpy.zeros((32,)))
            document_model.append_data_item(data_item)
            document_model.append_data_item(data_item2, False)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.display_type = "line_plot"
            display_item.append_display_data_channel(DisplayItem.DisplayDataChannel(data_item=data_item2))
            display_item.calibration_style_id = "pixels-top-left"
            p, v = display_item.get_value_and_position_text(display_item.display_data_channel, (25, ))
            self.assertEqual(p, "25")
            self.assertEqual(v, "")

    def test_cursor_over_1d_data_displays_without_exception(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((1000, )))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            display_panel.display_canvas_item.mouse_entered()
            display_panel.display_canvas_item.mouse_position_changed(500, 500, Graphics.NullModifiers())
            display_panel.display_canvas_item.mouse_exited()

    def test_cursor_over_1d_data_displays_without_exception_when_not_displaying_calibration(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((1000, )))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.calibration_style_id = "relative-top-left"
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            display_panel.display_canvas_item.mouse_entered()
            display_panel.display_canvas_item.mouse_position_changed(500, 500, Graphics.NullModifiers())
            display_panel.display_canvas_item.mouse_exited()

    def test_cursor_over_1d_multiple_data_displays_without_exception(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_and_metadata = DataAndMetadata.new_data_and_metadata(numpy.zeros((4, 1000), numpy.float64), data_descriptor=DataAndMetadata.DataDescriptor(False, 1, 1))
            data_item = DataItem.new_data_item(data_and_metadata)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.calibration_style_id = "pixels-top-left"
            p, v = display_item.get_value_and_position_text(display_item.display_data_channel, (500,))
            self.assertEqual(p, "500.0, 0.0")
            self.assertEqual(v, "0")

    def test_cursor_over_1d_multiple_data_but_2_datum_dimensions_displays_without_exception(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_and_metadata = DataAndMetadata.new_data_and_metadata(numpy.zeros((4, 1000), numpy.float64), data_descriptor=DataAndMetadata.DataDescriptor(False, 0, 2))
            data_item = DataItem.new_data_item(data_and_metadata)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.calibration_style_id = "pixels-top-left"
            p, v = display_item.get_value_and_position_text(display_item.display_data_channel, (500,))
            self.assertEqual(p, "500.0, 0.0")
            self.assertEqual(v, "0")

    def test_cursor_over_1d_sequence_data_displays_without_exception(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_and_metadata = DataAndMetadata.new_data_and_metadata(numpy.zeros((4, 1000), numpy.float64), data_descriptor=DataAndMetadata.DataDescriptor(True, 0, 1))
            data_item = DataItem.new_data_item(data_and_metadata)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.calibration_style_id = "pixels-top-left"
            p, v = display_item.get_value_and_position_text(display_item.display_data_channel, (500,))
            self.assertEqual(p, "500.0, 0.0")
            self.assertEqual(v, "0")

    def test_cursor_over_1d_image_without_exception(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.DataItem(numpy.zeros((50,)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.calibration_style_id = "pixels-top-left"
            p, v = display_item.get_value_and_position_text(display_item.display_data_channel, (25, ))
            self.assertEqual(p, "25.0")
            self.assertEqual(v, "0")

    def test_cursor_over_1d_image_without_exception_x(self):
        with TestContext.create_memory_context() as test_context:
            document_model = test_context.create_document_model()
            data_item = DataItem.new_data_item(DataAndMetadata.new_data_and_metadata(numpy.zeros((4, 25)), data_descriptor=DataAndMetadata.DataDescriptor(False, 1, 1)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.display_type = "image"
            p, v = display_item.get_value_and_position_text(display_item.display_data_channel, (2, 20))
            self.assertEqual(p, "20.0, 2.0")
            self.assertEqual(v, "0")

    def test_cursor_over_3d_data_displays_without_exception(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.zeros((10, 10, 4)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            display_panel.display_canvas_item.mouse_entered()
            display_panel.display_canvas_item.mouse_position_changed(500, 500, Graphics.NullModifiers())
            display_panel.display_canvas_item.mouse_exited()

    def test_cursor_over_3d_data_displays_correct_ordering_of_indices(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_item = DataItem.DataItem(numpy.ones((100, 100, 20)))
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.calibration_style_id = "pixels-top-left"
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            info_panel = document_controller.find_dock_panel("info-panel")
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            display_panel.display_canvas_item.mouse_entered()
            display_panel.display_canvas_item.mouse_position_changed(500, 500, Graphics.NullModifiers())
            document_controller.periodic()
            self.assertEqual(info_panel.label_row_1.text, "Position: 0.0, 50.0, 50.0")
            self.assertEqual(info_panel.label_row_2.text, "Value: 1")
            self.assertIsNone(info_panel.label_row_3.text, None)
            display_panel.display_canvas_item.mouse_exited()

    def test_cursor_over_2d_data_sequence_displays_correct_ordering_of_indices(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data_and_metadata = DataAndMetadata.new_data_and_metadata(numpy.ones((20, 100, 100), numpy.float64), data_descriptor=DataAndMetadata.DataDescriptor(True, 0, 2))
            data_item = DataItem.new_data_item(data_and_metadata)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.display_data_channels[0].sequence_index = 4
            display_item.calibration_style_id = "pixels-top-left"
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            info_panel = document_controller.find_dock_panel("info-panel")
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            display_panel.display_canvas_item.mouse_entered()
            document_controller.periodic()
            display_panel.display_canvas_item.mouse_position_changed(500, 500, Graphics.NullModifiers())
            document_controller.periodic()
            self.assertEqual(info_panel.label_row_1.text, "Position: 50.0, 50.0, 4.0")
            self.assertEqual(info_panel.label_row_2.text, "Value: 1")
            self.assertIsNone(info_panel.label_row_3.text, None)
            display_panel.display_canvas_item.mouse_exited()

    def test_cursor_over_4d_data_displays_correctly(self):
        with TestContext.create_memory_context() as test_context:
            document_controller = test_context.create_document_controller()
            document_model = document_controller.document_model
            display_panel = document_controller.selected_display_panel
            data = (numpy.random.randn(100, 100, 20, 20) * 100).astype(numpy.int32)
            data_item = DataItem.DataItem(data)
            document_model.append_data_item(data_item)
            display_item = document_model.get_display_item_for_data_item(data_item)
            display_item.calibration_style_id = "pixels-top-left"
            display_item.display_data_channels[0].collection_index = 20, 30
            display_panel.set_display_panel_display_item(display_item)
            header_height = display_panel.header_canvas_item.header_height
            info_panel = document_controller.find_dock_panel("info-panel")
            display_panel.root_container.layout_immediate((1000 + header_height, 1000))
            display_panel.display_canvas_item.mouse_entered()
            display_panel.display_canvas_item.mouse_position_changed(400, 600, Graphics.NullModifiers())
            document_controller.periodic()
            self.assertEqual(info_panel.label_row_1.text, "Position: 8.0, 12.0, 30.0, 20.0")
            self.assertEqual(info_panel.label_row_2.text, "Value: {}".format(data[20, 30, 12, 8]))
            self.assertIsNone(info_panel.label_row_3.text, None)
            display_panel.display_canvas_item.mouse_exited()



# standard libraries
import contextlib
import unittest

# third party libraries
# None

# local libraries
from nion.swift.model import Persistence


class TestPersistentObjectContextClass(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_persistent_object_context_does_not_trigger_event_on_already_registered_object(self):
        persistent_object_context = Persistence.PersistentObjectContext()
        object1 = Persistence.PersistentObject()
        persistent_object_context.register(object1, object1.item_specifier)
        was_registered = False
        def registered(registered_item, unregistered_item) -> None:
            nonlocal was_registered
            if registered_item:
                was_registered = True
        with persistent_object_context.registration_event.listen(registered):
            self.assertFalse(was_registered)

    def test_persistent_object_context_calls_register_when_object_becomes_registered(self):
        persistent_object_context = Persistence.PersistentObjectContext()
        object1 = Persistence.PersistentObject()
        was_registered = False
        def registered(registered_item, unregistered_item) -> None:
            nonlocal was_registered
            if registered_item == object1:
                was_registered = True
        with persistent_object_context.registration_event.listen(registered):
            persistent_object_context.register(object1, object1.item_specifier)
            self.assertTrue(was_registered)

    def test_persistent_object_context_calls_unregister_when_object_becomes_unregistered(self):
        persistent_object_context = Persistence.PersistentObjectContext()
        object1 = Persistence.PersistentObject()
        was_registered = False
        def registered(registered_item, unregistered_item) -> None:
            nonlocal was_registered
            if registered_item == object1:
                was_registered = True
            if unregistered_item == object1:
                was_registered = False
        with persistent_object_context.registration_event.listen(registered):
            item_specifier1 = object1.item_specifier
            persistent_object_context.register(object1, item_specifier1)
            self.assertTrue(was_registered)
            persistent_object_context.unregister(object1, item_specifier1)
            self.assertFalse(was_registered)

    def test_persistent_object_context_unregister_without_subscription_works(self):
        # this test will only generate extra output in the failure case, which has been fixed
        persistent_object_context = Persistence.PersistentObjectContext()
        object1 = Persistence.PersistentObject()
        persistent_object_context.register(object1, object1.item_specifier)
        object1 = None

    def test_persistent_object_proxy_updates_when_registered(self):
        persistent_object_context = Persistence.PersistentObjectContext()
        object0 = Persistence.PersistentObject()
        object1 = Persistence.PersistentObject()
        object1_proxy = object0.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(object1.uuid))
        with contextlib.closing(object1_proxy):
            object0.persistent_object_context = persistent_object_context
            self.assertIsNone(object1_proxy.item)
            object1.persistent_object_context = persistent_object_context
            self.assertEqual(object1, object1_proxy.item)

    def test_persistent_object_proxy_updates_when_unregistered(self):
        persistent_object_context = Persistence.PersistentObjectContext()
        object0 = Persistence.PersistentObject()
        object1 = Persistence.PersistentObject()
        object1_proxy = object0.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(object1.uuid))
        with contextlib.closing(object1_proxy):
            object0.persistent_object_context = persistent_object_context
            object1.persistent_object_context = persistent_object_context
            self.assertEqual(object1, object1_proxy.item)
            object1.persistent_object_context = None
            self.assertIsNone(object1_proxy.item)

    def test_persistent_object_proxy_calls_listeners_when_initialized_with_object(self):

        r_count = 0
        u_count = 0

        def r(x): nonlocal r_count; r_count += 1
        def u(x): nonlocal u_count; u_count += 1

        persistent_object_context = Persistence.PersistentObjectContext()
        object0 = Persistence.PersistentObject()
        object1 = Persistence.PersistentObject()
        object1_proxy = object0.create_item_proxy(item=object1)
        object1_proxy.on_item_registered = r
        object1_proxy.on_item_unregistered = u
        with contextlib.closing(object1_proxy):
            # register parent and check initial conditions
            object0.persistent_object_context = persistent_object_context
            self.assertEqual(object1, object1_proxy.item)
            self.assertEqual(0, r_count)
            self.assertEqual(0, u_count)
            # register child which was already registered; confirm registered not called
            object1.persistent_object_context = persistent_object_context
            self.assertEqual(0, r_count)  # item never becomes registered; it starts registered
            self.assertEqual(0, u_count)
            # unregister works normal
            object1.persistent_object_context = None
            self.assertEqual(0, r_count)
            self.assertEqual(1, u_count)

    def test_persistent_object_proxy_calls_listeners_once_during_lifecycle(self):

        r_count = 0
        u_count = 0

        def r(x): nonlocal r_count; r_count += 1
        def u(x): nonlocal u_count; u_count += 1

        persistent_object_context = Persistence.PersistentObjectContext()
        object0 = Persistence.PersistentObject()
        object1 = Persistence.PersistentObject()
        object1_proxy = object0.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(object1.uuid))
        with contextlib.closing(object1_proxy):
            object1_proxy.on_item_registered = r
            object1_proxy.on_item_unregistered = u
            # register parent, but not child
            object0.persistent_object_context = persistent_object_context
            self.assertEqual(0, r_count)
            self.assertEqual(0, u_count)
            # now register child and ensure register was called
            object1.persistent_object_context = persistent_object_context
            self.assertEqual(object1, object1_proxy.item)
            self.assertEqual(1, r_count)
            self.assertEqual(0, u_count)
            # unregister child and ensure unregistered was called
            object1.persistent_object_context = None
            self.assertEqual(1, r_count)
            self.assertEqual(1, u_count)

    def test_persistent_object_proxy_not_registered_when_already_registered(self):

        r_count = 0
        u_count = 0

        def r(x): nonlocal r_count; r_count += 1
        def u(x): nonlocal u_count; u_count += 1

        persistent_object_context = Persistence.PersistentObjectContext()
        object0 = Persistence.PersistentObject()
        object1 = Persistence.PersistentObject()
        object0.persistent_object_context = persistent_object_context
        object1.persistent_object_context = persistent_object_context
        # both objects are already registered
        object1_proxy = object0.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(object1.uuid))
        with contextlib.closing(object1_proxy):
            object1_proxy.on_item_registered = r
            object1_proxy.on_item_unregistered = u
            # registered should not be called because it was already registered
            self.assertEqual(object1, object1_proxy.item)
            self.assertEqual(0, r_count)
            self.assertEqual(0, u_count)
            object1.persistent_object_context = None
            # but unregistered should be called
            self.assertEqual(0, r_count)
            self.assertEqual(1, u_count)

    def test_persistent_object_proxy_does_not_call_listener_when_removed_from_unregistered_parent(self):

        r_count = 0
        u_count = 0

        def r(x): nonlocal r_count; r_count += 1
        def u(x): nonlocal u_count; u_count += 1

        persistent_object_context = Persistence.PersistentObjectContext()
        object0 = Persistence.PersistentObject()
        object1 = Persistence.PersistentObject()
        object1_proxy = object0.create_item_proxy(item_specifier=Persistence.PersistentObjectSpecifier.read(object1.uuid))
        with contextlib.closing(object1_proxy):
            object1_proxy.on_item_registered = r
            object1_proxy.on_item_unregistered = u
            object0.persistent_object_context = persistent_object_context
            # only parent has been registered
            self.assertEqual(0, r_count)
            self.assertEqual(0, u_count)
            object1.persistent_object_context = persistent_object_context
            # now object has been registered
            self.assertEqual(object1, object1_proxy.item)
            self.assertEqual(1, r_count)
            self.assertEqual(0, u_count)
            # unregistered parent first, then object
            object0.persistent_object_context = None
            object1.persistent_object_context = None
            self.assertEqual(1, r_count)
            self.assertEqual(0, u_count)  # parent was already unregistered



"""
A collection of persistence classes.
"""
from __future__ import annotations

from __future__ import annotations

# standard libraries
import abc
import copy
import datetime
import logging
import numpy
import re
import typing
import uuid
import weakref


# third party libraries
# None

# local libraries
from nion.utils import Event


class PersistentProperty:

    """
        Represents a persistent property.

        converter converts from value to json value
    """

    def __init__(self, name, value=None, make=None, read_only=False, hidden=False, recordable=True, validate=None, converter=None, changed=None, key=None, reader=None, writer=None):
        super().__init__()
        self.name = name
        self.key = key if key else name
        self.value = value
        self.make = make
        self.read_only = read_only
        self.hidden = hidden
        self.recordable = recordable
        self.validate = validate
        self.converter = converter
        self.reader = reader
        self.writer = writer
        self.convert_get_fn = converter.convert if converter else copy.deepcopy  # optimization
        self.convert_set_fn = converter.convert_back if converter else lambda value: value  # optimization
        self.changed = changed

    def close(self) -> None:
        self.make = None
        self.validate = None
        self.converter = None
        self.reader = None
        self.writer = None
        self.convert_get_fn = None
        self.convert_set_fn = None
        self.changed = None

    def set_value(self, value):
        if self.validate:
            value = self.validate(value)
        else:
            value = copy.deepcopy(value)
        self.value = value
        if self.changed:
            self.changed(self.name, value)

    @property
    def json_value(self):
        return self.convert_get_fn(self.value)

    @json_value.setter
    def json_value(self, json_value):
        self.set_value(self.convert_set_fn(json_value))

    def read_from_dict(self, properties):
        if self.reader:
            value = self.reader(self, properties)
            if value is not None:
                self.set_value(value)
        else:
            if self.key in properties:
                if self.make:
                    value = self.make()
                    value.read_dict(properties[self.key])
                    self.set_value(value)
                else:
                    self.json_value = properties[self.key]

    def write_to_dict(self, properties):
        if self.writer:
            self.writer(self, properties, self.value)
        else:
            if self.make:
                value = self.value
                if value is not None:
                    value_dict = value.write_dict()
                    properties[self.key] = value_dict
                else:
                    properties.pop(self.key, None)  # remove key
            else:
                value = self.json_value
                if value is not None:
                    properties[self.key] = value
                else:
                    properties.pop(self.key, None)  # remove key


class PersistentPropertySpecial(PersistentProperty):

    def __init__(self, name, value=None, make=None, read_only=False, hidden=False, recordable=True, validate=None, converter=None, changed=None, key=None, reader=None, writer=None):
        super().__init__(name, value, make, read_only, hidden, recordable, validate, converter, changed, key, reader, writer)
        self.__value = value

    @property
    def value(self):
        return copy.deepcopy(self.__value)

    @value.setter
    def value(self, value):
        self.__value = value


class PersistentItem:

    def __init__(self, name, factory, item_changed=None, hidden=False):
        super(PersistentItem, self).__init__()
        self.name = name
        self.factory = factory
        self.item_changed = item_changed
        self.hidden = hidden
        self.value = None

    def close(self) -> None:
        self.item_changed = None


class PersistentRelationship:

    def __init__(self, name, factory, insert=None, remove=None, key=None):
        super().__init__()
        self.name = name
        self.factory = factory
        self.insert = insert
        self.remove = remove
        self.key = key
        self.values = list()
        self.index = dict()

    def close(self) -> None:
        self.insert = None
        self.remove = None

    @property
    def storage_key(self):
        return self.key if self.key else self.name


class PersistentStorageInterface(abc.ABC):

    @abc.abstractmethod
    def get_storage_properties(self) -> typing.Dict: ...

    @abc.abstractmethod
    def get_properties(self, object) -> typing.Dict: ...

    @abc.abstractmethod
    def insert_item(self, parent, name: str, before_index: int, item) -> None: ...

    @abc.abstractmethod
    def remove_item(self, parent, name: str, index: int, item) -> None: ...

    @abc.abstractmethod
    def set_item(self, parent, name, item): ...

    @abc.abstractmethod
    def set_property(self, object, name, value): ...

    @abc.abstractmethod
    def clear_property(self, object, name): ...

    @abc.abstractmethod
    def read_external_data(self, item, name: str): ...

    @abc.abstractmethod
    def write_external_data(self, item, name: str, value) -> None: ...

    @abc.abstractmethod
    def reserve_external_data(self, item, name: str, data_shape: typing.Tuple[int, ...], data_dtype: numpy.dtype) -> None: ...

    @abc.abstractmethod
    def enter_write_delay(self, object) -> None: ...

    @abc.abstractmethod
    def exit_write_delay(self, object) -> None: ...

    @abc.abstractmethod
    def is_write_delayed(self, data_item) -> bool: ...

    @abc.abstractmethod
    def rewrite_item(self, item) -> None: ...


class PersistentObjectContext:
    """Provides a common context to track available persistent objects.

    All objects participating in this context should register and unregister themselves with this context at appropriate
    times.

    Other objects can listen to the registration_event to know when a specific object is registered or unregistered.
    """

    def __init__(self):
        self.__objects = dict()
        self.registration_event = Event.Event()
        self.__registration_changed_map: typing.Dict[uuid.UUID, typing.Dict[PersistentObjectProxy, typing.Callable[[typing.Optional[PersistentObject], typing.Optional[PersistentObject]], None]]] = dict()

    def unregister_registration_changed_fn(self, uuid_: uuid.UUID, key: PersistentObjectProxy) -> None:
        registration_changed_key_map = self.__registration_changed_map.get(uuid_, dict())
        registration_changed_key_map.pop(key)
        if not registration_changed_key_map:
            self.__registration_changed_map.pop(uuid_, None)

    def register_registration_changed_fn(self, uuid_: uuid.UUID, key: PersistentObjectProxy, registration_changed_fn: typing.Callable[[typing.Optional[PersistentObject], typing.Optional[PersistentObject]], None]):
        registration_changed_key_map = self.__registration_changed_map.setdefault(uuid_, dict())
        registration_changed_key_map[key] = registration_changed_fn

    def register(self, object: PersistentObject, item_specifier: PersistentObjectSpecifier) -> None:
        # print(f"register {object} {item_specifier.write()} {len(self.__objects) + 1}")
        # assert item_specifier not in self.__objects
        self.__objects[item_specifier] = weakref.ref(object)
        self.registration_event.fire(object, None)
        registration_changed_key_map = self.__registration_changed_map.get(object.uuid, dict())
        for registration_changed_fn in list(registration_changed_key_map.values()):
            if callable(registration_changed_fn):
                registration_changed_fn(object, None)

    def unregister(self, object: PersistentObject, item_specifier: PersistentObjectSpecifier) -> None:
        # print(f"unregister {object} {item_specifier.write()} {len(self.__objects) - 1}")
        # assert item_specifier in self.__objects
        if item_specifier in self.__objects:
            self.__objects.pop(item_specifier)
            self.registration_event.fire(None, object)
            registration_changed_key_map = self.__registration_changed_map.get(object.uuid, dict())
            for registration_changed_fn in list(registration_changed_key_map.values()):
                if callable(registration_changed_fn):
                    registration_changed_fn(None, object)

    def get_registered_object(self, item_specifier: PersistentObjectSpecifier) -> typing.Optional[PersistentObject]:
        object_weakref = self.__objects.get(item_specifier, None)
        return object_weakref() if object_weakref else None


class PersistentObjectSpecifier:

    def __init__(self, *, item: PersistentObject = None, item_uuid: uuid.UUID = None, context: PersistentObject = None, context_uuid: uuid.UUID = None):
        self.__item_uuid = item.uuid if item else item_uuid
        self.__context_uuid = context.uuid if context else context_uuid
        assert (self.__item_uuid is None) or isinstance(self.__item_uuid, uuid.UUID)

    def __hash__(self):
        return hash((self.__context_uuid, self.__item_uuid))

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__item_uuid == other.__item_uuid and self.__context_uuid == other.__context_uuid
        return False

    @property
    def item_uuid(self) -> typing.Optional[uuid.UUID]:
        return self.__item_uuid

    @property
    def context_uuid(self) -> typing.Optional[uuid.UUID]:
        return self.__context_uuid

    def write(self) -> typing.Optional[typing.Union[typing.Dict, str]]:
        if self.__item_uuid:
            if not self.__context_uuid:
                return str(self.__item_uuid)
            else:
                return {"item_uuid": str(self.__item_uuid), "context_uuid": str(self.__context_uuid)}
        return None

    @staticmethod
    def read(d: typing.Union[typing.Mapping, str, uuid.UUID]) -> typing.Optional[PersistentObjectSpecifier]:
        if isinstance(d, str):
            return PersistentObjectSpecifier(item_uuid=uuid.UUID(d))
        elif isinstance(d, uuid.UUID):
            return PersistentObjectSpecifier(item_uuid=d)
        elif isinstance(d, dict) and "item_uuid" in d and "context_uuid" in d:
            return PersistentObjectSpecifier(item_uuid=uuid.UUID(d["item_uuid"]), context_uuid=uuid.UUID(d["context_uuid"]))
        elif isinstance(d, dict) and "item_uuid" in d:
            return PersistentObjectSpecifier(item_uuid=uuid.UUID(d["item_uuid"]))
        elif isinstance(d, dict) and "uuid" in d:
            return PersistentObjectSpecifier(item_uuid=uuid.UUID(d["uuid"]), context_uuid=uuid.UUID(d["context_uuid"]) if "context_uuid" in d else None)
        else:
            return None


class PersistentObjectProxy:

    def __init__(self, persistent_object: PersistentObject, item_specifier: typing.Optional[PersistentObjectSpecifier], item: typing.Optional[PersistentObject]):
        self.__persistent_object = persistent_object
        self.__item_specifier = item_specifier if item_specifier else PersistentObjectSpecifier(item=item) if item else None
        self.__item = item
        self.__registration_listener = None
        self.__persistent_object_context = None
        self.__registered_change_uuid = None
        self.on_item_registered = None
        self.on_item_unregistered = None

        self.__persistent_object_context_changed_listener = persistent_object.persistent_object_context_changed_event.listen(self.__persistent_object_context_changed)

        # watch for our context being removed
        self.__persistent_object_about_to_be_removed_listener = persistent_object.about_to_be_removed_event.listen(self.close)

        self.__persistent_object_context_changed()

    def close(self):
        if self.__registration_listener:
            self.__registration_listener.close()
            self.__registration_listener = None
        if self.__persistent_object_context_changed_listener:
            self.__persistent_object_context_changed_listener.close()
            self.__persistent_object_context_changed_listener = None
        if self.__persistent_object_about_to_be_removed_listener:
            self.__persistent_object_about_to_be_removed_listener.close()
            self.__persistent_object_about_to_be_removed_listener = None
        self.__item = None
        self.__item_specifier = None
        self.__update_persistent_object_context()
        self.__persistent_object = None
        self.on_item_registered = None
        self.on_item_unregistered = None

    @property
    def item(self) -> typing.Optional[PersistentObject]:
        return self.__item

    @item.setter
    def item(self, item: typing.Optional[PersistentObject]) -> None:
        if self.__registration_listener:
            self.__registration_listener.close()
            self.__registration_listener = None
        self.__item = item
        self.__item_specifier = PersistentObjectSpecifier(item=item) if item else None
        self.__persistent_object_context_changed()

    @property
    def item_specifier(self) -> PersistentObjectSpecifier:
        return self.__item_specifier

    @item_specifier.setter
    def item_specifier(self, item_specifier: PersistentObjectSpecifier) -> None:
        if self.__registration_listener:
            self.__registration_listener.close()
            self.__registration_listener = None
        self.__item_specifier = item_specifier
        self.__item = None
        self.__persistent_object_context_changed()

    def __update_persistent_object_context(self):
        if self.__persistent_object_context:
            self.__persistent_object_context.unregister_registration_changed_fn(self.__registered_change_uuid, self)
            self.__registered_change_uuid = None
            self.__persistent_object_context = None
        if self.__item_specifier and self.__persistent_object.persistent_object_context:
            self.__persistent_object_context = self.__persistent_object.persistent_object_context
            self.__registered_change_uuid = self.__item_specifier.item_uuid
            self.__persistent_object_context.register_registration_changed_fn(self.__item_specifier.item_uuid, self, self.__change_registration)

    def __change_registration(self, registered_object: typing.Optional[PersistentObject], unregistered_object: typing.Optional[PersistentObject]) -> None:
        if registered_object and not self.__item and self.__item_specifier and registered_object.uuid == self.__item_specifier.item_uuid:
            item = self.__persistent_object._get_related_item(self.__item_specifier)
            if item:
                self.__item = item
                if callable(self.on_item_registered):
                    self.on_item_registered(registered_object)
        if unregistered_object and unregistered_object == self.__item:
            self.__item = None
            if callable(self.on_item_unregistered):
                self.on_item_unregistered(unregistered_object)

    def __persistent_object_context_changed(self) -> None:
        if self.__persistent_object.persistent_object_context:
            if self.__item_specifier and not self.__item:
                item = self.__persistent_object._get_related_item(self.__item_specifier)
                if item:
                    self.__change_registration(item, None)
        self.__update_persistent_object_context()


class PersistentObjectParent:
    """ Track the parent of a persistent object. """

    def __init__(self, parent: PersistentObject, relationship_name: str=None, item_name: str=None):
        self.__weak_parent = weakref.ref(parent)
        self.relationship_name = relationship_name
        self.item_name = item_name

    @property
    def parent(self) -> PersistentObject:
        return self.__weak_parent()


class PersistentObject:
    """
        Base class for objects being stored in a PersistentObjectContext.

        Subclasses can define properties, items, and relationships. Changes to those items will
        be persisted to the PersistentObjectContext.

        Keeps track of modified field automatically.

        Properties are single values, items are a one-to-one between objects, relationships are
        one-to-many between objects.

        Properties can have validators, converters, change notifications, and more.
        They are created using the define_property method.

        Items have set notifications and more. They are created using the
        define_item method.

        Relationships have change notifications and more. They are created using
        the define_relationship method.

        Subclasses can set the uuid after init. It should not be changed at other times.

        The persistent_object_context property must be set explicitly for top level objects of the
        PersistentObjectContext. Objects contained in the items and relationships will be have their
        persistent_object_context managed when they are inserted into or removed from another persistent object.

        It is an error to add an object as an item or relationship more than once. Items can be
        removed and added again, though.

        Objects must be able to read from and write themselves to a dict.

        The persistent_object_context property will be valid after (but not during) reading.

        After reading, an object may immediately update itself to a newer version using the persistent object
        context.
    """

    def __init__(self):
        super().__init__()
        self.__type = None
        self.__properties = dict()
        self.__items = dict()
        self.__relationships = dict()
        self.__container_weak_ref = None
        self._closed = False
        self.about_to_close_event = Event.Event()
        self._about_to_be_removed = False
        self.about_to_be_removed_event = Event.Event()
        self._is_reading = False
        self.__persistent_object_context = None
        # uuid as a property is too slow, so make it direct
        self.uuid = uuid.uuid4()
        self.__modified_count = 0
        self.modified_state = 0
        self.__modified = datetime.datetime.utcnow()
        self.persistent_object_parent = None
        self.__persistent_dict = None
        self.__persistent_storage = None
        self.persistent_object_context_changed_event = Event.Event()

    def close(self) -> None:
        self.about_to_close_event.fire()
        assert not self._closed
        self._closed = True
        self.close_items()
        self.close_relationships()
        self.__container_weak_ref = None
        self.__persistent_object_context = None
        self.undefine_properties()
        self.undefine_items()
        self.undefine_relationships()

    def __deepcopy__(self, memo):
        deepcopy = self.__class__()
        deepcopy.deepcopy_from(self, memo)
        memo[id(self)] = deepcopy
        return deepcopy

    def deepcopy_from(self, item, memo):
        for key in self.__properties.keys():
            value = item._get_persistent_property_value(key)
            new_value = copy.deepcopy(value)
            self._set_persistent_property_value(key, new_value)
        for key in self.__items.keys():
            self.set_item(key, copy.deepcopy(getattr(item, key)))
        for key in self.__relationships.keys():
            for child_item in getattr(item, key):
                self.append_item(key, copy.deepcopy(child_item, memo))

    @property
    def container(self):
        return self.__container_weak_ref() if self.__container_weak_ref else None

    def about_to_be_inserted(self, container):
        assert self.__container_weak_ref is None
        self.__container_weak_ref = weakref.ref(container)

    def about_to_be_removed(self, container):
        # called before close and before item is removed from its container
        for item in self.__items.values():
            item.value.about_to_be_removed(self)
        for relationship in self.__relationships.values():
            for item in reversed(relationship.values):
                item.about_to_be_removed(self)
        self.about_to_be_removed_event.fire()
        assert not self._about_to_be_removed
        self._about_to_be_removed = True
        self.__container_weak_ref = None

    @property
    def item_specifier(self) -> PersistentObjectSpecifier:
        return PersistentObjectSpecifier(item_uuid=self.uuid)

    def define_root_context(self) -> None:
        """Define this item to be the root context."""
        self.__persistent_object_context = PersistentObjectContext()

    def set_storage_system(self, storage_system: PersistentStorageInterface) -> None:
        """Set the storage system for this item."""
        self.persistent_dict = storage_system.get_storage_properties()
        self.persistent_storage = storage_system

    def update_storage_system(self) -> None:
        """Update the storage system properties by re-reading from storage.

        Useful when reloading.
        """
        self.persistent_dict = self.persistent_storage.get_storage_properties()

    def update_item_context(self, item: PersistentObject) -> None:
        """Update the context on the item."""
        item.persistent_object_context = self.persistent_object_context

    def persistent_object_context_changed(self) -> None:
        """ Subclasses can override this to be notified when the persistent object context changes. """
        pass

    @property
    def persistent_object_context(self) -> PersistentObjectContext:
        """ Return the persistent object context. """
        return self.__persistent_object_context

    @persistent_object_context.setter
    def persistent_object_context(self, persistent_object_context: PersistentObjectContext) -> None:
        """ Set the persistent object context and propagate it to contained objects. """
        assert self.__persistent_object_context is None or persistent_object_context is None  # make sure persistent object context is handled cleanly
        old_persistent_object_context = self.__persistent_object_context
        self.__persistent_object_context = persistent_object_context
        for item in self.__items.values():
            if item.value:
                item.value.persistent_object_context = persistent_object_context
        for relationship in self.__relationships.values():
            for item in relationship.values:
                if item:
                    item.persistent_object_context = persistent_object_context
        if old_persistent_object_context:
            old_persistent_object_context.unregister(self, self.__item_specifier)
        if persistent_object_context:
            self.__item_specifier = self.item_specifier
            persistent_object_context.register(self, self.__item_specifier)
        self.persistent_object_context_changed()
        self.persistent_object_context_changed_event.fire()

    @property
    def persistent_dict(self) -> typing.Dict:
        return self.__persistent_dict

    @persistent_dict.setter
    def persistent_dict(self, persistent_dict: typing.Dict) -> None:
        self.__persistent_dict = persistent_dict
        for key in self.__items.keys():
            item = self.__items[key].value
            if item:
                item.persistent_dict = self._get_item_persistent_dict(item, key) if persistent_dict is not None else None
                item.persistent_storage = self.persistent_storage if persistent_dict is not None else None
        for key in self.__relationships.keys():
            for index, item in enumerate(self.__relationships[key].values):
                item.persistent_dict = self._get_relationship_persistent_dict(item, key, index) if persistent_dict is not None else None
                item.persistent_storage = self.persistent_storage if persistent_dict is not None else None

    @property
    def persistent_storage(self) -> PersistentStorageInterface:
        return self.__persistent_storage

    @persistent_storage.setter
    def persistent_storage(self, persistent_storage: PersistentStorageInterface) -> None:
        self.__persistent_storage = persistent_storage
        for key in self.__items.keys():
            item = self.__items[key].value
            if item:
                item.persistent_storage = persistent_storage
        for key in self.__relationships.keys():
            for index, item in enumerate(self.__relationships[key].values):
                item.persistent_storage = persistent_storage

    def _get_item_persistent_dict(self, item, key: str) -> typing.Optional[typing.Dict]:
        return self.persistent_dict[key] if self.persistent_dict is not None else None

    def _get_relationship_persistent_dict(self, item, key: str, index: int) -> typing.Optional[typing.Dict]:
        return self.persistent_dict[key][index] if self.persistent_dict is not None else None

    def _get_relationship_persistent_dict_by_uuid(self, item, key: str) -> typing.Optional[typing.Dict]:
        if self.persistent_dict:
            item_uuid = str(item.uuid)
            for item_d in self.persistent_dict.get(key, list()):
                # if uuid.UUID(item_d.get("uuid")) == item.uuid:
                #     return item_d
                if item_d.get("uuid") == item_uuid:  # a little dangerous, comparing the uuid str's, significantly faster
                    return item_d
        return None

    def _get_related_item(self, item_specifier: PersistentObjectSpecifier) -> typing.Optional[PersistentObject]:
        if self.persistent_object_parent and self.persistent_object_parent.parent:
            return self.persistent_object_parent.parent._get_related_item(item_specifier)
        if self.persistent_object_context:
            return self.persistent_object_context.get_registered_object(item_specifier)
        return None

    def define_type(self, type):
        self.__type = type

    def define_property(self, name: str, value=None, make=None, read_only: bool=False, hidden: bool=False, recordable: bool=True, copy_on_read: bool=False, validate=None, converter=None, changed=None, key=None, reader=None, writer=None):
        """ key is what is stored on disk; name is what is used when accessing the property from code. """
        if copy_on_read:
            self.__properties[name] = PersistentPropertySpecial(name, value, make, read_only, hidden, recordable, validate, converter, changed, key, reader, writer)
        else:
            self.__properties[name] = PersistentProperty(name, value, make, read_only, hidden, recordable, validate, converter, changed, key, reader, writer)

    def define_item(self, name, factory, item_changed=None, hidden=False):
        self.__items[name] = PersistentItem(name, factory, item_changed, hidden)

    def define_relationship(self, name, factory, insert=None, remove=None, key=None):
        self.__relationships[name] = PersistentRelationship(name, factory, insert, remove, key)

    def close_items(self) -> None:
        for item in self.__items.values():
            if item.value:
                item.value.persistent_object_context = None
                item.value.close()

    def close_relationships(self) -> None:
        for relationship in self.__relationships.values():
            for item in reversed(relationship.values):
                if item:
                    item.persistent_object_context = None
                    item.close()

    def undefine_properties(self) -> None:
        for property in self.__properties.values():
            property.close()
        self.__properties.clear()

    def undefine_items(self) -> None:
        for item in self.__items.values():
            item.close()
        self.__items.clear()

    def undefine_relationships(self) -> None:
        for relationship in self.__relationships.values():
            relationship.close()
        self.__relationships.clear()

    def get_storage_properties(self) -> typing.Dict:
        """ Return a copy of the properties for the object as a dict. """
        return copy.deepcopy(self.persistent_storage.get_properties(self))

    @property
    def property_names(self):
        return list(self.__properties.keys())

    @property
    def key_names(self):
        return [property.key for property in self.__properties.values()]

    @property
    def type(self):
        return self.__type

    @property
    def modified(self):
        return self.__modified

    @property
    def modified_count(self):
        return self.__modified_count

    def _set_modified(self, modified):
        # for testing
        self.__update_modified(modified)
        if self.persistent_object_context:
            self.property_changed("uuid", str(self.uuid))  # dummy write

    @property
    def item_names(self):
        return list(self.__items.keys())

    @property
    def relationship_names(self):
        return list(self.__relationships.keys())

    def begin_reading(self):
        self._is_reading = True

    def read_from_dict(self, properties):
        """ Read from a dict. """
        # uuid is handled specially for performance reasons
        if "uuid" in properties:
            self.uuid = uuid.UUID(properties["uuid"])
        if "modified" in properties:
            self.__modified = datetime.datetime(*list(map(int, re.split('[^\d]', properties["modified"]))))
        # iterate the defined properties
        for key in self.__properties.keys():
            property = self.__properties[key]
            property.read_from_dict(properties)
        for key in self.__items.keys():
            item_dict = properties.get(key)
            if item_dict:
                factory = self.__items[key].factory
                # the object has not been constructed yet, but we needs its
                # type or id to construct it. so we need to look it up by key/index/name.
                # to minimize the interface to the factory methods, just pass a closure
                # which looks up by name.
                def lookup_id(name, default=None):
                    return item_dict.get(name, default)
                item = factory(lookup_id)
                if item is None:
                    logging.debug("Unable to read %s", key)
                assert item is not None
                # read the item from the dict
                item.begin_reading()
                item.read_from_dict(item_dict)
                self.__set_item(key, item)
                item.persistent_dict = self._get_item_persistent_dict(item, key)
                item.persistent_storage = self.persistent_storage
        for key in self.__relationships.keys():
            storage_key = self.__relationships[key].storage_key
            for item_dict in properties.get(storage_key, list()):
                factory = self.__relationships[key].factory
                # the object has not been constructed yet, but we needs its
                # type or id to construct it. so we need to look it up by key/index/name.
                # to minimize the interface to the factory methods, just pass a closure
                # which looks up by name.
                def lookup_id(name, default=None):
                    return item_dict.get(name, default)
                item = factory(lookup_id)
                if item is None:
                    logging.debug("Unable to read %s", key)
                assert item is not None
                # read the item from the dict
                item.begin_reading()
                item.read_from_dict(item_dict)
                # insert it into the relationship dict
                before_index = len(self.__relationships[key].values)
                self.load_item(key, before_index, item)

    def finish_reading(self):
        for key in self.__items.keys():
            item = self.__items[key].value
            if item:
                item.finish_reading()
        for key in self.__relationships.keys():
            for item in self.__relationships[key].values:
                item.finish_reading()
        self._is_reading = False

    def write_to_dict(self):
        """ Write the object to a dict and return it. """
        properties = dict()
        if self.__type:
            properties["type"] = self.__type
        properties["uuid"] = str(self.uuid)
        for key in self.__properties.keys():
            property = self.__properties[key]
            property.write_to_dict(properties)
        for key in self.__items.keys():
            item = self.__items[key].value
            if item:
                properties[key] = item.write_to_dict()
        for key in self.__relationships.keys():
            storage_key = self.__relationships[key].storage_key
            items_list = properties.setdefault(storage_key, list())
            for item in self.__relationships[key].values:
                items_list.append(item.write_to_dict())
        return properties

    def _update_persistent_object_context_property(self, name):
        """
            Update the property given by name in the persistent object context.

            Subclasses can override this to provide custom writing behavior, such
            as delaying write until an appropriate time for performance reasons.
        """
        if self.persistent_object_context:
            properties = dict()
            self.__properties[name].write_to_dict(properties)
            if properties:
                for property_key, property_value in properties.items():
                    self.property_changed(property_key, property_value)
            else:
                self.clear_property(name)

    def __update_modified(self, modified):
        self.__modified_count += 1
        self.modified_state += 1
        self.__modified = modified
        parent = self.persistent_object_parent.parent if self.persistent_object_parent else None
        if parent:
            parent.__update_modified(modified)

    def _get_persistent_property(self, name):
        """ Subclasses can call this to get a property descriptor. """
        return self.__properties[name]

    def _get_persistent_property_value(self, name, default=None):
        """ Subclasses can call this to get a hidden property. """
        property = self.__properties.get(name)
        return property.value if property else default

    def _set_persistent_property_value(self, name, value):
        """ Subclasses can call this to set a hidden property. """
        property = self.__properties[name]
        property.set_value(value)
        self.__update_modified(datetime.datetime.utcnow())
        self._update_persistent_object_context_property(name)

    def _update_persistent_property(self, name: str, value) -> None:
        """ Subclasses can call this to notify that a custom property was updated. """
        self.__update_modified(datetime.datetime.utcnow())
        if self.persistent_object_context:
            self.property_changed(name, value)

    def _is_persistent_property_recordable(self, name) -> bool:
        property = self.__properties.get(name)
        return (property.recordable and not property.read_only) if (property is not None) else False

    def __getattr__(self, name):
        # Handle property objects that are not hidden.
        property = self.__properties.get(name)
        if property and not property.hidden:
            return property.value
        if name in self.__items and not self.__items[name].hidden:
            return self.__items[name].value
        if name in self.__relationships:
            return copy.copy(self.__relationships[name].values)
        raise AttributeError("%r object has no attribute %r" % (self.__class__, name))

    def __setattr__(self, name, value):
        # Check for private properties of this class
        if name.startswith("_PersistentObject__"):
            super().__setattr__(name, value)
        # Otherwise check for defined properties.
        else:
            property = self.__properties.get(name)
            # if the property is hidden, fall through and give regular style property a chance to handle it
            if property and not property.hidden:
                # if the property is not hidden and it is read only, throw an exception
                if not property.read_only:
                    property.set_value(value)
                    self.__update_modified(datetime.datetime.utcnow())
                    self._update_persistent_object_context_property(name)
                else:
                    raise AttributeError()
            else:
                super().__setattr__(name, value)

    def __set_item(self, name, value):
        """ Set item into item storage and notify. Does not set into persistent storage or update modified. Item can be None. """
        item = self.__items[name]
        old_value = item.value
        item.value = value
        if value:
            value.persistent_object_parent = PersistentObjectParent(self, item_name=name)
            value.persistent_object_context = self.persistent_object_context
        if item.item_changed:
            item.item_changed(name, old_value, value)

    def get_item(self, name):
        """ Get item from persistent storage. """
        item = self.__items[name]
        return item.value

    def set_item(self, name, value):
        """ Set item into persistent storage and then into item storage and notify. """
        item = self.__items[name]
        old_value = item.value
        item.value = value
        self.__update_modified(datetime.datetime.utcnow())
        if value:
            value.persistent_object_parent = PersistentObjectParent(self, item_name=name)
        # the persistent_object_parent and item need to be established before
        # calling item_changed.
        if self.persistent_object_context:
            self.item_set(name, value)  # this will also update item's persistent_object_context
        if value:
            value.persistent_object_context = self.persistent_object_context
        else:
            value.persistent_object_context = None
        if item.item_changed:
            item.item_changed(name, old_value, value)

    def load_item(self, name: str, before_index: int, item: PersistentObject) -> None:
        """ Load item in persistent storage and then into relationship storage, but don't update modified or notify persistent storage. """
        item.persistent_dict = self._get_relationship_persistent_dict_by_uuid(item, name)
        item.persistent_storage = self.persistent_storage
        relationship = self.__relationships[name]
        relationship.values.insert(before_index, item)
        relationship.index[item.uuid] = item
        item.about_to_be_inserted(self)
        item.persistent_object_parent = PersistentObjectParent(self, relationship_name=name)
        item.persistent_object_context = self.persistent_object_context
        if relationship.insert:
            relationship.insert(name, before_index, item)

    def unload_item(self, name: str, index: int) -> None:
        """ Unload item from relationship storage and persistent storage, but don't update modified or notify persistent storage. """
        relationship = self.__relationships[name]
        item = relationship.values.pop(index)
        relationship.index.pop(item.uuid)
        item.about_to_be_removed(self)
        if relationship.remove:
            relationship.remove(name, index, item)
        item.persistent_object_context = None
        item.persistent_object_parent = None
        item.persistent_storage = None
        item.persistent_dict = None
        item.close()

    def insert_item(self, name, before_index, item):
        """ Insert item in persistent storage and then into relationship storage and notify. """
        relationship = self.__relationships[name]
        relationship.values.insert(before_index, item)
        relationship.index[item.uuid] = item
        self.__update_modified(datetime.datetime.utcnow())
        item.persistent_object_parent = PersistentObjectParent(self, relationship_name=name)
        # the persistent_object_parent and relationship need to be established before
        # calling item_inserted.
        item.about_to_be_inserted(self)
        if self.persistent_object_context:
            self.item_inserted(name, before_index, item)  # this will also update item's persistent_object_context
            item.persistent_object_context = self.persistent_object_context
        if relationship.insert:
            relationship.insert(name, before_index, item)

    def append_item(self, name, item):
        """ Append item and append to persistent storage. """
        self.insert_item(name, len(self.__relationships[name].values), item)

    def remove_item(self, name, item):
        """ Remove item and remove from persistent storage. """
        item.about_to_be_removed(self)
        relationship = self.__relationships[name]
        item_index = relationship.values.index(item)
        relationship.values.remove(item)
        relationship.index.pop(item.uuid)
        self.__update_modified(datetime.datetime.utcnow())
        if relationship.remove:
            relationship.remove(name, item_index, item)
        item.persistent_object_context = None
        if self.persistent_object_context:
            self.item_removed(name, item_index, item)  # this will also update item's persistent_object_context
        item.persistent_object_parent = None
        item.close()

    def extend_items(self, name, items):
        """ Append multiple items and add to persistent storage. """
        for item in items:
            self.append_item(name, item)

    def item_count(self, name: str) -> int:
        """Return the count of items in the relationship specified by name."""
        relationship = self.__relationships[name]
        return len(relationship.values)

    def item_index(self, name: str, item: object) -> int:
        """Return the index of item within the relationship specified by name."""
        relationship = self.__relationships[name]
        return relationship.values.index(item)

    def get_item_by_uuid(self, name: str, uuid: uuid.UUID) -> typing.Optional[PersistentObject]:
        """Return the item from the index by uuid."""
        relationship = self.__relationships[name]
        return relationship.index.get(uuid)

    def item_inserted(self, name: str, before_index: int, item: PersistentObject) -> None:
        """ Call this to notify this context that the item before before_index has just been inserted into the parent in
        the relationship with the given name. """
        self.persistent_storage.insert_item(self, name, before_index, item)

    def item_removed(self, name: str, index: int, item: PersistentObject) -> None:
        """ Call this to notify this context that the item at item_index has been removed from the parent in the
        relationship with the given name. """
        self.persistent_storage.remove_item(self, name, index, item)

    def item_set(self, name: str, item: PersistentObject) -> None:
        """ Call this to notify this context that an item with name has been set on the parent. """
        self.persistent_storage.set_item(self, name, item)

    def property_changed(self, name: str, value) -> None:
        """ Call this to notify this context that a property with name has changed to value on object. """
        self.persistent_storage.set_property(self, name, value)

    def clear_property(self, name: str) -> None:
        """ Call this to notify this context that a property with name has been removed on object. """
        self.persistent_storage.clear_property(self, name)

    def read_external_data(self, name: str):
        """ Call this to notify read external data with name from an item in persistent storage. """
        return self.persistent_storage.read_external_data(self, name)

    def write_external_data(self, name: str, value) -> None:
        """ Call this to notify write external data value with name to an item in persistent storage. """
        self.persistent_storage.write_external_data(self, name, value)

    def reserve_external_data(self, name: str, data_shape: typing.Tuple[int, ...], data_dtype: numpy.dtype) -> None:
        """ Call this to notify reserve external data value with name to an item in persistent storage. """
        self.persistent_storage.reserve_external_data(self, name, data_shape, data_dtype)

    def enter_write_delay(self) -> None:
        """ Call this to notify this context that the object should be write delayed. """
        self.persistent_storage.enter_write_delay(self)

    def exit_write_delay(self) -> None:
        """ Call this to notify this context that the object should no longer be write delayed. """
        self.persistent_storage.exit_write_delay(self)

    @property
    def is_write_delayed(self) -> bool:
        return self.persistent_storage.is_write_delayed(self)

    def rewrite(self) -> None:
        """ Call this to write an item that was write delayed. """
        self.persistent_storage.rewrite_item(self)

    def create_item_proxy(self, *, item_uuid: uuid.UUID = None, item_specifier: PersistentObjectSpecifier = None, item: PersistentObject = None) -> PersistentObjectProxy:
        """Create an item proxy by uuid or directly using the item."""
        item_specifier = item_specifier or (PersistentObjectSpecifier(item_uuid=item_uuid) if item_uuid else None)
        return PersistentObjectProxy(self, item_specifier, item)

    def resolve_item_specifier(self, item_specifier: PersistentObjectSpecifier) -> typing.Optional[PersistentObject]:
        """Return the resolve item specifier."""
        return self._get_related_item(item_specifier)
